#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model import ModelFunction
from pinecone.protos import core_pb2
from pinecone.utils import load_numpy, clear_pb_repeated

import numpy as np

from typing import Iterable


class Scorer(ModelFunction):

    def score(self,
              data_queries: Iterable[np.ndarray],
              data_items: Iterable[Iterable[np.ndarray]]
              ) -> Iterable[Iterable[float]]:
        """
        Use a model to predictions scores. Each query is associated with an Iterable of items.
        The model makes a prediction for each query-item combination.
        :param data_queries:
        :param data_items: input array to model
        :return: numpy vector representing embedding
        """
        raise NotImplementedError

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        """

        @type msg: object
        """
        query_vectors = load_numpy(msg.query.data)
        all_scores = self.score(query_vectors,(load_numpy(matches.data) for matches in msg.query.matches) )
        for query_vec, matches, scores in zip(query_vectors, msg.query.matches, all_scores):
            clear_pb_repeated(matches.scores)
            matches.scores.extend(scores)
        return msg



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Iterable
import numpy as np

from pinecone.utils import load_numpy
from pinecone.grpc import GRPCClient

from pinecone.functions.model import HubFunction
from . import QueryResult


class HubPostprocessor(HubFunction):
    """Useful for local testing. Mimics a query request sent to a HubFunction hosting a Postprocessor image."""

    async def transform(self, queries: np.ndarray, matches: Iterable[QueryResult], timeout: float = 30) -> Iterable[QueryResult]:
        grpc = GRPCClient()
        msg = grpc.get_query_request(data=queries, matches=[mat._asdict() for mat in matches], path="read")
        response = await self.handle_msg(msg, timeout)

        result = []
        for match in response.query.matches:
            result.append(
                QueryResult(ids=match.ids, scores=match.scores, data=load_numpy(match.data) if match.data else None)
            )
        return result



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import List, NamedTuple
from pinecone.specs import traffic_router as traffic_router_spec

from pinecone.utils.sentry import sentry_decorator as sentry
from .api_router import RouterAPI
from .api_controller import ControllerAPI
from .constants import Config

__all__ = ["describe", "deploy", "stop", "ls", "update_active_service", "update_services", "RouterMeta"]


class RouterMeta(NamedTuple):
    """Metadata of a traffic router."""

    name: str
    active_service: str
    services: List[str]
    status: dict


def _get_router_api():
    return RouterAPI(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)


def _get_controller_api():
    return ControllerAPI(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)


@sentry
def ls():
    """Returns the names of all traffic routers."""
    api = _get_router_api()
    return api.list()


@sentry
def deploy(router_name: str, services: List[str], active_service: str = None):
    """Creates a traffic router.

    By default the first service will be set as active unless otherwise specified.

    :param router_name: name of the traffic router
    :type router_name: str
    :param services: the list of services that the traffic router can route traffic to.
        By default, the first service will be set as the active service.
    :type services: List[str]
    :param active_service: defaults to None. if specified, the given service will be the active service for the traffic router.
    :type active_service: str, optional
    """
    if active_service and active_service not in services:
        raise RuntimeError("{active_service} must be one of the services {services}".format(**locals()))

    active_service = active_service or services[0]

    controller_api = _get_controller_api()
    router_api = _get_router_api()

    user_services = controller_api.list_services()
    if set(services) - set(user_services):
        raise RuntimeError("These services are not yet available: {0}".format(list(set(services) - set(user_services))))

    router_ = traffic_router_spec.TrafficRouter(router_name, services, active_service)
    return router_api.deploy(router_.to_json())


@sentry
def update_active_service(router_name: str, active_service: str):
    """Updates the active service of a router.

    :param router_name: name of the traffic router
    :type router_name: str
    :param active_service: name of the active service. It has to be a service already in a traffic router.
    :type active_service: str
    """
    api = _get_router_api()
    router_json = api.get_router(router_name)
    router_ = traffic_router_spec.TrafficRouter.from_json(router_json)
    if active_service not in router_.all_services:
        raise RuntimeError("Service {0} is not in router {1}".format(active_service, router_.name))

    router_.active_service = active_service
    return api.deploy(router_.to_json())


@sentry
def update_services(router_name: str, services: List[str]):
    """Updates the list of services available in the traffic router.

    :param router_name: name of the traffic router
    :type router_name: str
    :param services: the list of services that the traffic router can route traffic to.
    :type services: List[str]
    """
    controller_api = _get_controller_api()
    router_api = _get_router_api()

    user_services = controller_api.list_services()
    if set(services) - set(user_services):
        raise RuntimeError("These services are not yet available: {0}".format(list(set(services) - set(user_services))))

    router_json = router_api.get_router(router_name)
    router_ = traffic_router_spec.TrafficRouter.from_json(router_json)
    if router_.active_service not in services:
        raise RuntimeError(
            "The new services must include the current active service {0}.".format(router_.active_service)
        )

    router_.all_services = services
    return router_api.deploy(router_.to_json())


@sentry
def stop(router_name: str):
    """Stops a router.

    :param router_name: the name of the router
    """
    api = _get_router_api()
    return api.stop(router_name)


@sentry
def describe(router_name: str):
    """Returns the metadata of a router.

    :param router_name: the name of the router
    """
    api = _get_router_api()
    router_json = api.get_router(router_name)
    router_ = traffic_router_spec.TrafficRouter.from_json(router_json)
    return RouterMeta(
        name=router_.name,
        active_service=router_.active_service,
        services=router_.all_services,
        status=api.get_status(router_name) or {},
    )



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model.transformer import Transformer

import numpy as np


class NormalizeTransformer(Transformer):
    """
    Normalize vectors (l2-norm)
    """
    def transform_vectors(self, vectors: np.ndarray) -> np.ndarray:
        data_array = np.array(vectors)
        data_norm = np.sqrt(np.einsum('ij,ij->i',data_array, data_array))
        col_dim = data_array.shape[1]
        return data_array / np.broadcast_to(data_norm, (col_dim,) + data_norm.shape).T



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import numpy as np

from pinecone.utils import load_numpy, dump_numpy
from pinecone.protos import core_pb2

from pinecone.functions.model import ModelFunction


class Preprocessor(ModelFunction):
    def transform(self, vectors: np.ndarray) -> np.ndarray:
        """Transforms vectors."""
        raise NotImplementedError

    def handle_msg(self, msg: "core_pb2.Request") -> "core_pb2.Request":
        req_type = msg.WhichOneof("body")
        if req_type == "index":
            vectors = load_numpy(msg.index.data)
            msg.index.data.CopyFrom(dump_numpy(self.transform(vectors)))
        elif req_type == "query":
            vectors = load_numpy(msg.query.data)
            msg.query.data.CopyFrom(dump_numpy(self.transform(vectors)))
        return msg


from pinecone.protos import core_pb2
from pinecone.utils import constants
from pinecone.utils import get_hostname, replica_kube_hostname, replica_name, tracing
from pinecone.network.zmq.spec import ServletSpec, SocketType, Socket
from pinecone.network.zmq.socket_wrapper import SocketWrapper

from loguru import logger
from typing import List, Union
from collections import defaultdict
import asyncio
import zmq
import zmq.asyncio

from ddtrace import tracer
from ddtrace.tracer import Context


class ZMQServlet:

    def __init__(self, servlet_spec: ServletSpec):
        self.spec = servlet_spec
        self.context = zmq.asyncio.Context()

        self.exception = None
        if len(self.spec.in_sockets) > 0:
            self.zmq_ins = [self.init_socket(in_sock) for in_sock in self.spec.in_sockets]
        self.zmq_outs = {r: [self.init_socket(s) for s in sockets] for r, sockets in self.spec.out_sockets.items()}

        self.msg_sent = 0
        self.msg_recv = 0
        self.replica_outs = defaultdict(list)  # Dict[function_name, List[socks to each replica]]
        self.gateway_control = {}

    def gateway_control_sock(self, replica: int) -> SocketWrapper:
        if replica not in self.gateway_control:
            self.gateway_control[replica] = self.init_socket(Socket(False, SocketType.PUSH,
                                                                    constants.ZMQ_CONTROL_PORT,
                                                                    host=self.get_host(constants.GATEWAY_NAME, replica)))
        return self.gateway_control[replica]

    @property
    def pretty_name(self) -> str:
        return self.spec.function_name + " shard:" + str(self.spec.shard) + " replica:" + str(self.spec.replica)

    async def handle_msg(self, msg: core_pb2.Request):
        route = core_pb2.Route(function=get_hostname(), function_id=self.spec.shard)
        route.start_time.GetCurrentTime()

        msg_context = Context(trace_id=msg.telemetry_trace_id, span_id=msg.telemetry_parent_id)
        tracer.context_provider.activate(msg_context)
        with tracer.trace(self.spec.function_name) as span:
            tracing.set_span_tags(span, msg)
            msg.telemetry_parent_id = span.span_id
            if self.spec.handle_msg:
                response = await self.spec.handle_msg(msg)
                route.end_time.GetCurrentTime()
                msg.routes.append(route)
                if response:
                    if self.spec.shard != 0:
                        response.shard_num = self.spec.shard
                    await self.send_msg(response)
                    del response
        del msg

    async def poll_sock(self, sock: SocketWrapper):
        loop = asyncio.get_event_loop()
        while True:
            msg = await self.recv_msg(sock)
            loop.create_task(self.handle_msg(msg))

    def start_polling(self) -> List[asyncio.Task]:
        loop = asyncio.get_event_loop()
        return [loop.create_task(self.poll_sock(sock)) for sock in self.zmq_ins]

    def refresh_dns(self) -> List[asyncio.Task]:
        loop = asyncio.get_event_loop()
        return [loop.create_task(sock.refresh_dns()) for sock in self.all_socks]

    def get_host(self, svc_name: str, replica: int):
        return replica_name(svc_name, 0, replica) if self.spec.native \
            else replica_kube_hostname(svc_name, 0, replica)

    def get_replica_sock(self, replica_num: int, sock: Socket) -> SocketWrapper:
        while replica_num >= len(self.replica_outs[sock.host]):
            new_sock = Socket(sock.bind, sock_type=sock.sock_type, port=sock.port,
                              host=self.get_host(sock.host, len(self.replica_outs[sock.host])))
            self.replica_outs[sock.host].append(self.init_socket(new_sock, disable_loadbalance=True))
        return self.replica_outs[sock.host][replica_num]

    async def send_msg(self, msg: 'core_pb2.Request'):
        self.msg_sent += 1
        send_sockets = []
        for path in {msg.path, '*'}:
            for sock in (sock for sock in self.spec.out_sockets.get(path, [])
                         if sock.host in [constants.GATEWAY_NAME, constants.AGGREGATOR_NAME]):
                send_sockets.append(self.get_replica_sock(msg.gateway_num, sock))
                break
            else:
                send_sockets.extend(self.zmq_outs.get(path, []))

        if len(send_sockets) == 0:
            logger.warning('{}: no out socket_spec for path {}'.format(get_hostname(), msg.path))

        msgs = self.spec.out_socket_selector.select_socket(msg, len(send_sockets))
        if len(msgs) == 1:
            serialized_msg = msgs[0].SerializeToString()
            await asyncio.gather(*(self.send_msg_to_single_socket(socket, serialized_msg) for socket in send_sockets))
        else:
            await asyncio.gather(*(self.send_msg_to_single_socket(socket, msg) for socket, msg in zip(send_sockets, msgs)))

        if msg.traceroute:
            receipt = core_pb2.TraceRoute(request_id=msg.request_id, client_id=msg.client_id,
                                          client_offset=msg.client_offset, routes=msg.routes)
            await self.gateway_control_sock(msg.gateway_num).send(receipt.SerializeToString())

    async def send_msg_to_single_socket(self, socket: SocketWrapper, msg: Union[bytes, core_pb2.Request]):
        if msg is None:
            return
        msg_bytes = msg if isinstance(msg, bytes) else msg.SerializeToString()
        await socket.send(msg_bytes)

    async def recv_msg(self, sock: SocketWrapper) -> 'core_pb2.Request':
        msg = await sock.recv()
        msg_pb = core_pb2.Request()
        msg_pb.ParseFromString(msg)
        self.msg_recv += 1
        return msg_pb

    def init_socket(self, socket: Socket, disable_loadbalance: bool = False):
        if socket.bind:
            logger.info(f"Listening on {socket.host}:{socket.port}")
        else:
            logger.info(f"Connecting to {socket.host}:{socket.port}")
        return SocketWrapper(socket, self.context, self.spec.native,
                             disable_loadbalance=disable_loadbalance or (socket.host and '.' in socket.host))

    @property
    def all_socks(self):
        return [*self.zmq_ins, *[sock for sock_list in self.zmq_outs.values() for sock in sock_list]]

    def cleanup(self):
        for sock in self.all_socks:
            sock.close()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function
from pinecone.specs import Spec
from pinecone.utils import validate_dns_name

from collections import defaultdict
from pydoc import locate
import argparse




class Service (Spec):
    """
    Class that represents a functional pipeline of vector search as a graph.
    """

    def __init__(self, name: str = None, gateway_replicas: int = 1):
        """
        :param gateway_port: Port to run the gRPC gateway on
        :param native: Whether to run with native python processes or kubernetes
        """
        self._name = name
        self.paths = defaultdict(list)
        self.paths['read'] = []
        self.paths['write'] = []
        self.functions = {}
        self.write_paths = {'write'}
        self.gateway_replicas = gateway_replicas

    @property
    def name(self) -> str:
        """Name of the service."""
        return self._name

    @name.setter
    def name(self, value: str):
        validate_dns_name(value)
        self._name = value

    def set_path_as_write(self, path_name: str):
        """
        Set a given path as a "write" path, used for stateful services
        :param path_name:
        :return:
        """
        self.write_paths.add(path_name)

    def unset_path_as_write(self, path_name: str):
        """
        Set a given path as not a "write" path, used for stateful functions
        :param path_name:
        :return:
        """
        self.write_paths.remove(path_name)

    def append_to_path(self, function: Function, path: str):
        """Appends a function to a given path.

        :param function: a Pinecone Function
        :type function: Function
        :param path: path name
        :type path: str
        """
        self.functions[function.name] = function
        self.paths[path].append(function.name)

    def append_to_all_paths(self, function: Function):
        """Appends a function to every path.

        :param function: a Pinecone Function
        :type function: Function
        """
        self.functions[function.name] = function
        self.paths = {k: [*v, function.name] for k, v in self.paths.items()}

    def remove_from_all_paths(self, name: str):
        """Appends a function from every path.

        :param name: name of Pinecone Function to remove
        :type name: str
        """
        if name in self.functions:
            del self.functions[name]
        self.paths = {k: [func for func in v if func != name] for k, v in self.paths.items()}

    def validate(self):
        """Validates the service.

        Perform the following validations:

        - There should be no orphaned nodes; that is, every node should belong to at least one path.
        - There should be no missing nodes; that is, any node name specified in one of the paths
          should have be in the list of functions.
        - Make sure that the paths have no cycles.
        - Service name and node names should be DNS compatible
        """
        node_names = [name for name in self.functions.keys()]
        all_steps = [ss for path in self.paths.values() for ss in path]
        orphaned_nodes = set(node_names) - set(all_steps)
        missing_nodes = set(all_steps) - set(node_names)

        for name in node_names:
            validate_dns_name(name)

        if orphaned_nodes:
            raise ValueError("Orphaned nodes (every node must belong to at least one path): {}".format(orphaned_nodes))
        if missing_nodes:
            raise ValueError("Missing nodes (at least one path has missing nodes): {}".format(missing_nodes))

        # Check for cycles
        for path_name, steps in self.paths.items():
            if len(steps) > len(set(steps)):
                raise ValueError("The following path contains cycles: ({}, {})".format(path_name, steps))

    def to_obj(self) -> dict:
        """Serializes specs."""
        metadata = {'name': self.name}
        spec = {
            'functions': [function.to_dict() for function in self.functions.values()],
            'paths': [{
                'name': name,
                'steps': steps
            } for name, steps in self.paths.items()],
            'write_paths': list(self.write_paths),
            'gateway_replicas': self.gateway_replicas
        }
        return {
            'version': 'pinecone/v1alpha1',
            'kind': 'Service',
            'metadata': metadata,
            'spec': spec
        }

    @classmethod
    def from_obj(cls, state_dict: dict) -> "Service":
        """Returns an instance of a Pinecone service from a graph.

        :return: :class:`Service`
        :rtype: :class:`Service`
        """
        assert state_dict['kind'] == 'Service'

        metadata = state_dict['metadata']
        spec = state_dict['spec']
        app = cls(metadata['name'], gateway_replicas=spec['gateway_replicas'])

        parser = argparse.ArgumentParser()
        Function.add_args(parser)
        app.functions = {
            function['name']: locate(function['class']).from_args(parser.parse_args(function['args']))
            for function in spec['functions']
        }
        app.paths = {path['name']: path['steps'] for path in spec['paths']}
        app.write_paths = set(spec['write_paths'])
        return app



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function
from pinecone.protos import core_pb2
from .hub import HubModel


class ModelFunction(Function):
    def setup(self):
        """
        Download and cache the model prior to serving
        :return:
        """


class HubFunction(HubModel, ModelFunction):
    def __init__(self, image: str, config: dict = None, **kwargs):
        """A special Pinecone Function that loads models from the Model Hub.

        :param image: a Pinecone Model Hub docker image URI
        :type image: str
        :param config: configurations for the model, defaults to None
        :type config: dict, optional
        """
        super().__init__(image=image, config=config, **kwargs)

    async def handle_msg(self, msg: "core_pb2.Request", timeout: float = 30) -> "core_pb2.Request":
        return await self.call_remote(msg, timeout)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.network.selector import SocketSelector
from pinecone.protos import core_pb2

from typing import List, Optional


class AllSelector(SocketSelector):

    @classmethod
    def select_socket(cls, orig_msg: 'core_pb2.Request', shards: int) -> List[Optional['core_pb2.Request']]:
        return [orig_msg]



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pathlib import Path
import inspect

from pinecone.protos import core_pb2
import numpy as np
from loguru import logger
import hashlib
import os
import re


DNS_COMPATIBLE_REGEX = re.compile("^[a-z0-9]([a-z0-9]|[-])+[a-z0-9]$")

def load_numpy(proto_arr: 'core_pb2.NdArray') -> 'np.ndarray':
    """
    Load numpy array from protobuf
    :param proto_arr:
    :return:
    """
    if len(proto_arr.shape) == 0:
        return np.array([])
    # if proto_arr.compressed:
    #     numpy_arr = np.frombuffer(lz4.frame.decompress(proto_arr.buffer), dtype=proto_arr.dtype)
    # else:
    numpy_arr = np.frombuffer(proto_arr.buffer, dtype=proto_arr.dtype)
    return numpy_arr.reshape(proto_arr.shape)


def dump_numpy(np_array: 'np.ndarray', compressed: bool = False) -> 'core_pb2.NdArray':
    """
    Dump numpy array to protobuf
    :param np_array:
    :param compressed: whether to use lz4 compression
    :return:
    """
    protobuf_arr = core_pb2.NdArray()
    protobuf_arr.dtype = str(np_array.dtype)
    protobuf_arr.shape.extend(np_array.shape)
    # if compressed:
    #     protobuf_arr.buffer = lz4.frame.compress(np_array.tobytes())
    #     protobuf_arr.compressed = True
    # else:
    protobuf_arr.buffer = np_array.tobytes()
    return protobuf_arr


HUB_CLASSES = set()


def hubify(cls):
    global HUB_CLASSES
    HUB_CLASSES.add(cls)
    return cls


def load_hub_service(module):
    global HUB_CLASSES
    subclasses = [getattr(module, d) for d in dir(module)
                  if inspect.isclass(getattr(module, d)) and getattr(module, d) in HUB_CLASSES]
    if len(subclasses) > 1:
        logger.error(f'Found more than one hub service, unable to load: {subclasses}.')
    elif len(subclasses) == 0:
        logger.error(f'No implementations of {HUB_CLASSES} found.')
    else:
        return subclasses[0]


def module_name(module):
    return module.__class__.__module__ + '.' + module.__class__.__name__


def get_version():
    return Path(__file__).parent.parent.joinpath('__version__').read_text().strip()


def get_environment():
    return Path(__file__).parent.parent.joinpath('__environment__').read_text().strip()


def clear_pb_repeated(field):
    while len(field) > 0:
        field.pop()


def get_native_port(name: str):
    return 6000 + int(hashlib.sha1(name.encode()).hexdigest(), 16) % 20000


def shard_name(name: str, shard_id: int):
    if shard_id > 0:
        return name + '-s' + str(shard_id)
    return name


def replica_from_shard_name(name: str, replica_id: int = None):
    if replica_id is not None:
        return name + '-' + str(replica_id)
    return name


def replica_name(name: str, shard_id: int, replica_id: int = None):  # shard == 0 is router
    svc_name = shard_name(name, shard_id)
    return replica_from_shard_name(svc_name, replica_id)


def replica_kube_hostname(name: str, shard_id: int, replica_id: int):
    return replica_name(name, shard_id, replica_id) + '.' + shard_name(name, shard_id)


def replica_kube_hostname_from_shard(name: str, replica_id: int):
    return replica_from_shard_name(name, replica_id) + '.' + name


def parse_hostname(function_name: str, hostname: str):
    hostname_list = hostname.replace(function_name, '').split('-')
    shard_id = None
    replica_id = None

    if 'deployment' in hostname:
        return shard_id, replica_id

    for item in hostname_list[1:]:
        if item[0] == 's':
            try:
                shard_id = int(item[1:])
            except ValueError:
                pass
        else:
            try:
                replica_id = int(item)
            except ValueError:
                pass
    return shard_id, replica_id


def open_or_create(path: str, truncate: int = None):
    if os.path.exists(path):
        file = open(path, 'r+b')
    else:
        file = open(path, 'w+b')
    if truncate:
        file.truncate(truncate)
    return file


def get_hostname():
    return os.environ.get("HOSTNAME")


def get_container_memory_usage():
    with open('/sys/fs/cgroup/memory/memory.stat') as stat_file:
        rss_line = stat_file.readlines()[1].strip()
        return int(rss_line.split()[1])


def get_container_memory_limit():
    MEMORY_LIMIT_FILE = '/sys/fs/cgroup/memory/memory.limit_in_bytes'
    if not os.path.exists(MEMORY_LIMIT_FILE):
        return 0  # no limit
    with open(MEMORY_LIMIT_FILE) as stat_file:
        return int(stat_file.read())


def validate_dns_name(name):
    if not DNS_COMPATIBLE_REGEX.match(name):
        raise ValueError("{} is invalid - service names and node names must consist of lower case "
                         "alphanumeric characters or '-', start with an alphabetic character, and end with an "
                         "alphanumeric character (e.g. 'my-name', or 'abc-123')".format(name))



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#
from pinecone.network.zmq.spec import Socket
from pinecone.utils import constants
import os
import random
import socket
import asyncio
from typing import Dict
import zmq
from loguru import logger
from functools import partial
from concurrent.futures import ThreadPoolExecutor

SOCK_ADDR = 4
FAMILY = 0
HOST = 0


def get_dns_hosts(socket_spec: Socket):
    try:
        addrinfo = socket.getaddrinfo(socket_spec.host, socket_spec.port)
        return list(set([sock[SOCK_ADDR][HOST] for sock in addrinfo if sock[FAMILY] == socket.AF_INET]))
    except socket.gaierror:
        return []


class SocketWrapper:

    def __init__(self, socket_spec: Socket, context: zmq.Context, native: bool, disable_loadbalance: bool = False):
        """
        :param socket:
        :param context:
        :param native:
        """
        self.context = context
        self.native = native
        self.socket_spec = socket_spec
        self.socket = None
        self.socket_pool = dict()  # type: Dict[str, zmq.Socket]
        self.socket_q = []  # array for round-robin
        self.num_sockets = 0
        self.sock_q_pos = 0
        self.unused_sockets = set()
        self.sockets_read_event = asyncio.Event()
        self.executor = ThreadPoolExecutor()
        if native or socket_spec.bind or disable_loadbalance:
            self.socket = self.get_socket(socket_spec)
        else:
            self.init_dns_sockets(socket_spec)

    def init_socket_q(self):
        self.socket_q = list(self.socket_pool.values())
        self.num_sockets = len(self.socket_q)
        if self.num_sockets > 0:
            self.sock_q_pos = random.randint(0, self.num_sockets -1)
            self.sockets_read_event.set()
        else:
            self.sockets_read_event.clear()

    def init_dns_sockets(self, socket_spec: Socket):
        hosts = get_dns_hosts(socket_spec)
        for ip in hosts:
            self.socket_pool[ip] = self.get_socket(self.socket_for_host(ip))

        self.init_socket_q()

    def get_conn_string(self, socket_spec: Socket) -> str:
        sock_path = '/tmp/'
        if socket_spec.bind:
            if self.native:
                ipc_sock = os.path.join(
                    sock_path, socket_spec.host) + str(socket_spec.port)

                if os.path.exists(ipc_sock):
                    os.remove(ipc_sock)

                conn_str = "ipc://{}".format(ipc_sock)
            else:
                conn_str = "tcp://*:{}".format(socket_spec.port)
        else:
            if self.native:
                ipc_sock = os.path.join(
                    sock_path, socket_spec.host) + str(socket_spec.port)
                conn_str = 'ipc://{}'.format(ipc_sock)
            else:
                conn_str = 'tcp://{}:{}'.format(socket_spec.host, socket_spec.port)
        return conn_str

    def socket_for_host(self, host: str) -> Socket:
        return Socket(self.socket_spec.bind, self.socket_spec.sock_type, self.socket_spec.port, host=host)

    def get_socket(self, socket_spec: Socket) -> zmq.Socket:
        if len(self.unused_sockets) > 0:
            zmq_socket, conn_str = self.unused_sockets.pop()
            zmq_socket.disconnect(conn_str)
        else:
            zmq_socket = socket_spec.zmq(self.context)
            zmq_socket.set_hwm(constants.MAX_MSGS_PENDING)  # needs to be set before bind/connect
            zmq_socket.setsockopt(zmq.RCVBUF, 2 * constants.MAX_MSG_SIZE)
            zmq_socket.setsockopt(zmq.SNDBUF, 2 * constants.MAX_MSG_SIZE)

        conn_str = self.get_conn_string(socket_spec)
        if socket_spec.bind:
            zmq_socket.bind(conn_str)
        else:
            zmq_socket.connect(conn_str)

        zmq_socket.setsockopt(zmq.LINGER, 0)
        return zmq_socket

    async def refresh_dns(self):
        loop = asyncio.get_event_loop()
        while not self.socket:  # only run for sockets that need dns load-balancing
            await asyncio.sleep(constants.DNS_TTL)
            new_dns = await loop.run_in_executor(self.executor, partial(get_dns_hosts, self.socket_spec))

            old_dns = self.socket_pool.keys()

            hosts_to_close = set(old_dns) - set(new_dns)
            hosts_to_open = set(new_dns) - set(old_dns)
            if hosts_to_close or hosts_to_open:
                logger.debug(f"closing {len(hosts_to_close)} sockets, opening {len(hosts_to_open)}")
            for entry in hosts_to_open:
                if len(hosts_to_close) > 0:
                    host_to_disconnect = hosts_to_close.pop()
                    sock_to_recycle = self.socket_pool.pop(host_to_disconnect)

                    sock_to_recycle.disconnect(self.get_conn_string(self.socket_for_host(host_to_disconnect)))
                    sock_to_recycle.connect(self.get_conn_string(self.socket_for_host(entry)))
                    self.socket_pool[entry] = sock_to_recycle
                else:
                    new_sock = self.get_socket(self.socket_for_host(entry))
                    self.socket_pool[entry] = new_sock

            for entry in hosts_to_close:
                sock_to_close = self.socket_pool.pop(entry)
                sock_to_close.disconnect(self.get_conn_string(self.socket_for_host(entry)))
                # connect to the host itself, so messages don't get dropped
                conn_str = self.get_conn_string(self.socket_for_host(self.socket_spec.host))
                sock_to_close.connect(conn_str)
                self.unused_sockets.add((sock_to_close, conn_str))
            self.init_socket_q()


    @property
    def next_socket(self):
        self.sock_q_pos = (self.sock_q_pos + 1) % self.num_sockets
        return self.socket_q[self.sock_q_pos]

    async def send(self, data: bytes):
        if self.socket:
            self.socket.send(data)
        else:
            await self.sockets_read_event.wait()
            await self.next_socket.send(data)

    async def recv(self) -> bytes:
        if self.socket:
            return await self.socket.recv()
        await self.sockets_read_event.wait()
        recv_socket = self.next_socket
        return await recv_socket.recv()

    def close(self):
        if self.socket:
            self.socket.close()
        for sock in self.socket_pool.values():
            if not sock.closed:
                sock.close()
        for sock, conn_str in self.unused_sockets:
            if not sock.closed:
                sock.close()

    @property
    def closed(self):
        if self.socket:
            return self.socket.closed
        return any([sock.closed for sock in self.socket_pool.values()])



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model.transformer import Transformer
from pinecone import utils


@utils.hubify
class NoOpTransformer(Transformer):
    def transform_vectors(self, vectors):
        return vectors



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: core.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='core.proto',
  package='core',
  syntax='proto3',
  serialized_options=b'\n\021io.pinecone.proto',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\ncore.proto\x12\x04\x63ore\x1a\x1fgoogle/protobuf/timestamp.proto\"m\n\x15ServiceControlRequest\x12\x10\n\x08\x66unction\x18\x01 \x01(\t\x12\x13\n\x0b\x66unction_id\x18\x02 \x01(\x04\x12\x1c\n\x06status\x18\x03 \x01(\x0b\x32\x0c.core.Status\x12\x0f\n\x07service\x18\x04 \x01(\t\"K\n\x07NdArray\x12\x0e\n\x06\x62uffer\x18\x01 \x01(\x0c\x12\r\n\x05shape\x18\x02 \x03(\r\x12\r\n\x05\x64type\x18\x03 \x01(\t\x12\x12\n\ncompressed\x18\x04 \x01(\x08\"\x8c\x01\n\x05Route\x12\x10\n\x08\x66unction\x18\x01 \x01(\t\x12\x13\n\x0b\x66unction_id\x18\x02 \x01(\x05\x12.\n\nstart_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\"\x91\x04\n\x07Request\x12\x12\n\nrequest_id\x18\x01 \x01(\x04\x12\x0f\n\x07timeout\x18\x02 \x01(\r\x12\x0c\n\x04path\x18\x03 \x01(\t\x12\x0f\n\x07version\x18\x04 \x01(\t\x12\x1b\n\x06routes\x18\x05 \x03(\x0b\x32\x0b.core.Route\x12\x1c\n\x06status\x18\x06 \x01(\x0b\x32\x0c.core.Status\x12#\n\x05query\x18\x07 \x01(\x0b\x32\x12.core.QueryRequestH\x00\x12#\n\x05index\x18\x08 \x01(\x0b\x32\x12.core.IndexRequestH\x00\x12%\n\x06\x64\x65lete\x18\t \x01(\x0b\x32\x13.core.DeleteRequestH\x00\x12!\n\x04info\x18\n \x01(\x0b\x32\x11.core.InfoRequestH\x00\x12#\n\x05\x66\x65tch\x18\x0b \x01(\x0b\x32\x12.core.FetchRequestH\x00\x12\x11\n\tnamespace\x18\x0c \x01(\t\x12\x11\n\tclient_id\x18\r \x01(\r\x12\x15\n\rclient_offset\x18\x0e \x01(\r\x12\x11\n\tshard_num\x18\x0f \x01(\r\x12\x13\n\x0bgateway_num\x18\x10 \x01(\r\x12\x1a\n\x12telemetry_trace_id\x18\x11 \x01(\x04\x12\x1b\n\x13telemetry_parent_id\x18\x12 \x01(\x04\x12\x14\n\x0cservice_name\x18\x13 \x01(\t\x12\x12\n\ntraceroute\x18\x14 \x01(\x08\x42\x06\n\x04\x62ody\"\xc3\x03\n\x06Status\x12%\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x17.core.Status.StatusCode\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12%\n\x07\x64\x65tails\x18\x03 \x03(\x0b\x32\x14.core.Status.Details\x12\x10\n\x08msg_sent\x18\x04 \x01(\x04\x12\x10\n\x08msg_recv\x18\x05 \x01(\x04\x12+\n\x08\x61vg_time\x18\x06 \x03(\x0b\x32\x19.core.Status.AvgTimeEntry\x12\x0c\n\x04size\x18\x07 \x01(\x04\x1a\x80\x01\n\x07\x44\x65tails\x12\x10\n\x08\x66unction\x18\x01 \x01(\t\x12\x13\n\x0b\x66unction_id\x18\x02 \x01(\t\x12\x11\n\texception\x18\x03 \x01(\t\x12\x11\n\ttraceback\x18\x04 \x01(\t\x12(\n\x04time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a.\n\x0c\x41vgTimeEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"D\n\nStatusCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\t\n\x05READY\x10\x01\x12\t\n\x05\x45RROR\x10\x02\x12\x13\n\x0f\x45RROR_DUPLICATE\x10\x03\"I\n\rScoredResults\x12\x0b\n\x03ids\x18\x01 \x03(\t\x12\x0e\n\x06scores\x18\x02 \x03(\x02\x12\x1b\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\r.core.NdArray\"8\n\x0cIndexRequest\x12\x0b\n\x03ids\x18\x01 \x03(\t\x12\x1b\n\x04\x64\x61ta\x18\x02 \x01(\x0b\x32\r.core.NdArray\"v\n\x0cQueryRequest\x12\r\n\x05top_k\x18\x01 \x01(\r\x12\x14\n\x0cinclude_data\x18\x02 \x01(\x08\x12\x1b\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\r.core.NdArray\x12$\n\x07matches\x18\x04 \x03(\x0b\x32\x13.core.ScoredResults\"\x1c\n\rDeleteRequest\x12\x0b\n\x03ids\x18\x01 \x03(\t\"!\n\x0bInfoRequest\x12\x12\n\nindex_size\x18\x01 \x01(\x04\";\n\x0c\x46\x65tchRequest\x12\x0b\n\x03ids\x18\x01 \x03(\t\x12\x1e\n\x07vectors\x18\x02 \x03(\x0b\x32\r.core.NdArray\"&\n\x03\x41\x63k\x12\x0f\n\x07replica\x18\x01 \x01(\r\x12\x0e\n\x06replay\x18\x02 \x01(\x08\"\\\n\x08LogEntry\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12\x1e\n\x05\x65ntry\x18\x02 \x01(\x0b\x32\r.core.RequestH\x00\x12\x18\n\x03\x61\x63k\x18\x03 \x01(\x0b\x32\t.core.AckH\x00\x42\x06\n\x04\x64\x61ta\"g\n\nTraceRoute\x12\x12\n\nrequest_id\x18\x01 \x01(\x04\x12\x11\n\tclient_id\x18\x02 \x01(\r\x12\x15\n\rclient_offset\x18\x03 \x01(\r\x12\x1b\n\x06routes\x18\x04 \x03(\x0b\x32\x0b.core.Route2d\n\tRPCClient\x12*\n\x04\x43\x61ll\x12\r.core.Request\x1a\r.core.Request\"\x00(\x01\x30\x01\x12+\n\tCallUnary\x12\r.core.Request\x1a\r.core.Request\"\x00\x42\x13\n\x11io.pinecone.protob\x06proto3'
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])



_STATUS_STATUSCODE = _descriptor.EnumDescriptor(
  name='StatusCode',
  full_name='core.Status.StatusCode',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='READY', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='ERROR_DUPLICATE', index=3, number=3,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1300,
  serialized_end=1368,
)
_sym_db.RegisterEnumDescriptor(_STATUS_STATUSCODE)


_SERVICECONTROLREQUEST = _descriptor.Descriptor(
  name='ServiceControlRequest',
  full_name='core.ServiceControlRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='core.ServiceControlRequest.function', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='function_id', full_name='core.ServiceControlRequest.function_id', index=1,
      number=2, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='core.ServiceControlRequest.status', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service', full_name='core.ServiceControlRequest.service', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=53,
  serialized_end=162,
)


_NDARRAY = _descriptor.Descriptor(
  name='NdArray',
  full_name='core.NdArray',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='buffer', full_name='core.NdArray.buffer', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shape', full_name='core.NdArray.shape', index=1,
      number=2, type=13, cpp_type=3, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='dtype', full_name='core.NdArray.dtype', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='compressed', full_name='core.NdArray.compressed', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=164,
  serialized_end=239,
)


_ROUTE = _descriptor.Descriptor(
  name='Route',
  full_name='core.Route',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='core.Route.function', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='function_id', full_name='core.Route.function_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='core.Route.start_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='end_time', full_name='core.Route.end_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=242,
  serialized_end=382,
)


_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='core.Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_id', full_name='core.Request.request_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='timeout', full_name='core.Request.timeout', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='path', full_name='core.Request.path', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='version', full_name='core.Request.version', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='routes', full_name='core.Request.routes', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='status', full_name='core.Request.status', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='query', full_name='core.Request.query', index=6,
      number=7, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='index', full_name='core.Request.index', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='delete', full_name='core.Request.delete', index=8,
      number=9, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='info', full_name='core.Request.info', index=9,
      number=10, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='fetch', full_name='core.Request.fetch', index=10,
      number=11, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='namespace', full_name='core.Request.namespace', index=11,
      number=12, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_id', full_name='core.Request.client_id', index=12,
      number=13, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_offset', full_name='core.Request.client_offset', index=13,
      number=14, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='shard_num', full_name='core.Request.shard_num', index=14,
      number=15, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='gateway_num', full_name='core.Request.gateway_num', index=15,
      number=16, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='telemetry_trace_id', full_name='core.Request.telemetry_trace_id', index=16,
      number=17, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='telemetry_parent_id', full_name='core.Request.telemetry_parent_id', index=17,
      number=18, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='service_name', full_name='core.Request.service_name', index=18,
      number=19, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='traceroute', full_name='core.Request.traceroute', index=19,
      number=20, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='body', full_name='core.Request.body',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=385,
  serialized_end=914,
)


_STATUS_DETAILS = _descriptor.Descriptor(
  name='Details',
  full_name='core.Status.Details',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='core.Status.Details.function', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='function_id', full_name='core.Status.Details.function_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='exception', full_name='core.Status.Details.exception', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='traceback', full_name='core.Status.Details.traceback', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='time', full_name='core.Status.Details.time', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1122,
  serialized_end=1250,
)

_STATUS_AVGTIMEENTRY = _descriptor.Descriptor(
  name='AvgTimeEntry',
  full_name='core.Status.AvgTimeEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='core.Status.AvgTimeEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='core.Status.AvgTimeEntry.value', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1252,
  serialized_end=1298,
)

_STATUS = _descriptor.Descriptor(
  name='Status',
  full_name='core.Status',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='code', full_name='core.Status.code', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='description', full_name='core.Status.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='details', full_name='core.Status.details', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='msg_sent', full_name='core.Status.msg_sent', index=3,
      number=4, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='msg_recv', full_name='core.Status.msg_recv', index=4,
      number=5, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='avg_time', full_name='core.Status.avg_time', index=5,
      number=6, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size', full_name='core.Status.size', index=6,
      number=7, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_STATUS_DETAILS, _STATUS_AVGTIMEENTRY, ],
  enum_types=[
    _STATUS_STATUSCODE,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=917,
  serialized_end=1368,
)


_SCOREDRESULTS = _descriptor.Descriptor(
  name='ScoredResults',
  full_name='core.ScoredResults',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ids', full_name='core.ScoredResults.ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='scores', full_name='core.ScoredResults.scores', index=1,
      number=2, type=2, cpp_type=6, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='core.ScoredResults.data', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1370,
  serialized_end=1443,
)


_INDEXREQUEST = _descriptor.Descriptor(
  name='IndexRequest',
  full_name='core.IndexRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ids', full_name='core.IndexRequest.ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='core.IndexRequest.data', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1445,
  serialized_end=1501,
)


_QUERYREQUEST = _descriptor.Descriptor(
  name='QueryRequest',
  full_name='core.QueryRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='top_k', full_name='core.QueryRequest.top_k', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='include_data', full_name='core.QueryRequest.include_data', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='data', full_name='core.QueryRequest.data', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='matches', full_name='core.QueryRequest.matches', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1503,
  serialized_end=1621,
)


_DELETEREQUEST = _descriptor.Descriptor(
  name='DeleteRequest',
  full_name='core.DeleteRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ids', full_name='core.DeleteRequest.ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1623,
  serialized_end=1651,
)


_INFOREQUEST = _descriptor.Descriptor(
  name='InfoRequest',
  full_name='core.InfoRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='index_size', full_name='core.InfoRequest.index_size', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1653,
  serialized_end=1686,
)


_FETCHREQUEST = _descriptor.Descriptor(
  name='FetchRequest',
  full_name='core.FetchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='ids', full_name='core.FetchRequest.ids', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='vectors', full_name='core.FetchRequest.vectors', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1688,
  serialized_end=1747,
)


_ACK = _descriptor.Descriptor(
  name='Ack',
  full_name='core.Ack',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='replica', full_name='core.Ack.replica', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='replay', full_name='core.Ack.replay', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1749,
  serialized_end=1787,
)


_LOGENTRY = _descriptor.Descriptor(
  name='LogEntry',
  full_name='core.LogEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='offset', full_name='core.LogEntry.offset', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='entry', full_name='core.LogEntry.entry', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='ack', full_name='core.LogEntry.ack', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='data', full_name='core.LogEntry.data',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=1789,
  serialized_end=1881,
)


_TRACEROUTE = _descriptor.Descriptor(
  name='TraceRoute',
  full_name='core.TraceRoute',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='request_id', full_name='core.TraceRoute.request_id', index=0,
      number=1, type=4, cpp_type=4, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_id', full_name='core.TraceRoute.client_id', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='client_offset', full_name='core.TraceRoute.client_offset', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='routes', full_name='core.TraceRoute.routes', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1883,
  serialized_end=1986,
)

_SERVICECONTROLREQUEST.fields_by_name['status'].message_type = _STATUS
_ROUTE.fields_by_name['start_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ROUTE.fields_by_name['end_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_REQUEST.fields_by_name['routes'].message_type = _ROUTE
_REQUEST.fields_by_name['status'].message_type = _STATUS
_REQUEST.fields_by_name['query'].message_type = _QUERYREQUEST
_REQUEST.fields_by_name['index'].message_type = _INDEXREQUEST
_REQUEST.fields_by_name['delete'].message_type = _DELETEREQUEST
_REQUEST.fields_by_name['info'].message_type = _INFOREQUEST
_REQUEST.fields_by_name['fetch'].message_type = _FETCHREQUEST
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['query'])
_REQUEST.fields_by_name['query'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['index'])
_REQUEST.fields_by_name['index'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['delete'])
_REQUEST.fields_by_name['delete'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['info'])
_REQUEST.fields_by_name['info'].containing_oneof = _REQUEST.oneofs_by_name['body']
_REQUEST.oneofs_by_name['body'].fields.append(
  _REQUEST.fields_by_name['fetch'])
_REQUEST.fields_by_name['fetch'].containing_oneof = _REQUEST.oneofs_by_name['body']
_STATUS_DETAILS.fields_by_name['time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_STATUS_DETAILS.containing_type = _STATUS
_STATUS_AVGTIMEENTRY.containing_type = _STATUS
_STATUS.fields_by_name['code'].enum_type = _STATUS_STATUSCODE
_STATUS.fields_by_name['details'].message_type = _STATUS_DETAILS
_STATUS.fields_by_name['avg_time'].message_type = _STATUS_AVGTIMEENTRY
_STATUS_STATUSCODE.containing_type = _STATUS
_SCOREDRESULTS.fields_by_name['data'].message_type = _NDARRAY
_INDEXREQUEST.fields_by_name['data'].message_type = _NDARRAY
_QUERYREQUEST.fields_by_name['data'].message_type = _NDARRAY
_QUERYREQUEST.fields_by_name['matches'].message_type = _SCOREDRESULTS
_FETCHREQUEST.fields_by_name['vectors'].message_type = _NDARRAY
_LOGENTRY.fields_by_name['entry'].message_type = _REQUEST
_LOGENTRY.fields_by_name['ack'].message_type = _ACK
_LOGENTRY.oneofs_by_name['data'].fields.append(
  _LOGENTRY.fields_by_name['entry'])
_LOGENTRY.fields_by_name['entry'].containing_oneof = _LOGENTRY.oneofs_by_name['data']
_LOGENTRY.oneofs_by_name['data'].fields.append(
  _LOGENTRY.fields_by_name['ack'])
_LOGENTRY.fields_by_name['ack'].containing_oneof = _LOGENTRY.oneofs_by_name['data']
_TRACEROUTE.fields_by_name['routes'].message_type = _ROUTE
DESCRIPTOR.message_types_by_name['ServiceControlRequest'] = _SERVICECONTROLREQUEST
DESCRIPTOR.message_types_by_name['NdArray'] = _NDARRAY
DESCRIPTOR.message_types_by_name['Route'] = _ROUTE
DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Status'] = _STATUS
DESCRIPTOR.message_types_by_name['ScoredResults'] = _SCOREDRESULTS
DESCRIPTOR.message_types_by_name['IndexRequest'] = _INDEXREQUEST
DESCRIPTOR.message_types_by_name['QueryRequest'] = _QUERYREQUEST
DESCRIPTOR.message_types_by_name['DeleteRequest'] = _DELETEREQUEST
DESCRIPTOR.message_types_by_name['InfoRequest'] = _INFOREQUEST
DESCRIPTOR.message_types_by_name['FetchRequest'] = _FETCHREQUEST
DESCRIPTOR.message_types_by_name['Ack'] = _ACK
DESCRIPTOR.message_types_by_name['LogEntry'] = _LOGENTRY
DESCRIPTOR.message_types_by_name['TraceRoute'] = _TRACEROUTE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ServiceControlRequest = _reflection.GeneratedProtocolMessageType('ServiceControlRequest', (_message.Message,), {
  'DESCRIPTOR' : _SERVICECONTROLREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.ServiceControlRequest)
  })
_sym_db.RegisterMessage(ServiceControlRequest)

NdArray = _reflection.GeneratedProtocolMessageType('NdArray', (_message.Message,), {
  'DESCRIPTOR' : _NDARRAY,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.NdArray)
  })
_sym_db.RegisterMessage(NdArray)

Route = _reflection.GeneratedProtocolMessageType('Route', (_message.Message,), {
  'DESCRIPTOR' : _ROUTE,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.Route)
  })
_sym_db.RegisterMessage(Route)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.Request)
  })
_sym_db.RegisterMessage(Request)

Status = _reflection.GeneratedProtocolMessageType('Status', (_message.Message,), {

  'Details' : _reflection.GeneratedProtocolMessageType('Details', (_message.Message,), {
    'DESCRIPTOR' : _STATUS_DETAILS,
    '__module__' : 'core_pb2'
    # @@protoc_insertion_point(class_scope:core.Status.Details)
    })
  ,

  'AvgTimeEntry' : _reflection.GeneratedProtocolMessageType('AvgTimeEntry', (_message.Message,), {
    'DESCRIPTOR' : _STATUS_AVGTIMEENTRY,
    '__module__' : 'core_pb2'
    # @@protoc_insertion_point(class_scope:core.Status.AvgTimeEntry)
    })
  ,
  'DESCRIPTOR' : _STATUS,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.Status)
  })
_sym_db.RegisterMessage(Status)
_sym_db.RegisterMessage(Status.Details)
_sym_db.RegisterMessage(Status.AvgTimeEntry)

ScoredResults = _reflection.GeneratedProtocolMessageType('ScoredResults', (_message.Message,), {
  'DESCRIPTOR' : _SCOREDRESULTS,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.ScoredResults)
  })
_sym_db.RegisterMessage(ScoredResults)

IndexRequest = _reflection.GeneratedProtocolMessageType('IndexRequest', (_message.Message,), {
  'DESCRIPTOR' : _INDEXREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.IndexRequest)
  })
_sym_db.RegisterMessage(IndexRequest)

QueryRequest = _reflection.GeneratedProtocolMessageType('QueryRequest', (_message.Message,), {
  'DESCRIPTOR' : _QUERYREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.QueryRequest)
  })
_sym_db.RegisterMessage(QueryRequest)

DeleteRequest = _reflection.GeneratedProtocolMessageType('DeleteRequest', (_message.Message,), {
  'DESCRIPTOR' : _DELETEREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.DeleteRequest)
  })
_sym_db.RegisterMessage(DeleteRequest)

InfoRequest = _reflection.GeneratedProtocolMessageType('InfoRequest', (_message.Message,), {
  'DESCRIPTOR' : _INFOREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.InfoRequest)
  })
_sym_db.RegisterMessage(InfoRequest)

FetchRequest = _reflection.GeneratedProtocolMessageType('FetchRequest', (_message.Message,), {
  'DESCRIPTOR' : _FETCHREQUEST,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.FetchRequest)
  })
_sym_db.RegisterMessage(FetchRequest)

Ack = _reflection.GeneratedProtocolMessageType('Ack', (_message.Message,), {
  'DESCRIPTOR' : _ACK,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.Ack)
  })
_sym_db.RegisterMessage(Ack)

LogEntry = _reflection.GeneratedProtocolMessageType('LogEntry', (_message.Message,), {
  'DESCRIPTOR' : _LOGENTRY,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.LogEntry)
  })
_sym_db.RegisterMessage(LogEntry)

TraceRoute = _reflection.GeneratedProtocolMessageType('TraceRoute', (_message.Message,), {
  'DESCRIPTOR' : _TRACEROUTE,
  '__module__' : 'core_pb2'
  # @@protoc_insertion_point(class_scope:core.TraceRoute)
  })
_sym_db.RegisterMessage(TraceRoute)


DESCRIPTOR._options = None
_STATUS_AVGTIMEENTRY._options = None

_RPCCLIENT = _descriptor.ServiceDescriptor(
  name='RPCClient',
  full_name='core.RPCClient',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=1988,
  serialized_end=2088,
  methods=[
  _descriptor.MethodDescriptor(
    name='Call',
    full_name='core.RPCClient.Call',
    index=0,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_REQUEST,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
  _descriptor.MethodDescriptor(
    name='CallUnary',
    full_name='core.RPCClient.CallUnary',
    index=1,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_REQUEST,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_RPCCLIENT)

DESCRIPTOR.services_by_name['RPCClient'] = _RPCCLIENT

# @@protoc_insertion_point(module_scope)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.engine import Engine
from pinecone.protos import core_pb2
from pydoc import locate
from pinecone.utils import logger
import json
import math


class NamespacedEngine(Engine):
    """
    This engine allows indexing of vectors in namespaces using exact or approximated engines, such that
    each namespace will is handled by a unique sub-engine.
    """

    SUPPORTED_ENGINES = {'approximated': 'pinecone_engine.functions.engine.approximated.ApproximatedEngine',
                         'exact': 'pinecone_engine.functions.engine.exact.ExactEngine',
                         'exactly': 'pinecone_engine.functions.engine.exact.ExactlyEngine'}
    SUPPORTED_ENGINE_IMAGES = {'approximated': 'pinecone/engine/approximated',
                               'exact': 'pinecone/engine/exact',
                               'exactly': 'pinecone/engine/exactly'}

    @property
    def image(self):
        return self.SUPPORTED_ENGINE_IMAGES[self.config['engine_type']]

    def __init__(self, engine_type="approximated", *args,  **config):
        """
        NamespacedEngine Constructor
        :param engine_type: type of engine to use for each namespace.
                            The supported engines are 'approximated' and 'exact':
                            'approximated' -  This engine allows a fast approximated search based on common metrics
                                              It's based on Pinecone's proprietary vector index
                            'exact' -  This engine allows an accurate vector search based on common metrics
                                       It's based on an exhaustive search, thus might be slow
        :param metric: type of metric used in the vector index -
                        'cosine' - cosine similarity
                        'dotproduct' - dot-product
                        'euclidean' - euclidean distance
        :return:
        """
        super().__init__(*args, **config)
        self.config['engine_type'] = engine_type
        self.config['engine_cpus'] = math.ceil(self.cpu_request / 1000.0)
        self.engine_class = None
        self.engines = {}

    def setup(self):
        super().setup()
        self.engine_class = self.get_engine_class()
        self.export_metadata()

    def cleanup(self):
        for engine in self.engines.values():
            engine.cleanup()

    def get_engine_class(self):
        engine_type = self.config['engine_type']
        if engine_type not in NamespacedEngine.SUPPORTED_ENGINES:
            raise ValueError(f'{engine_type} is not a supported engine type')
        return locate(NamespacedEngine.SUPPORTED_ENGINES[engine_type])

    def export_metadata(self):
        logger.opt(raw=True).info(json.dumps({"recordType": 'namespacedEngine.metadata',
                                              "name": self.name, **self.config,
                                              "namespaces": list(self.engines.keys())}) + "\n")

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        namespace = msg.namespace
        req_type = msg.WhichOneof('body')
        if req_type == 'index':
            if namespace not in self.engines:
                engine_name = f'{self.name}-{self.id}-{namespace}'
                new_engine = self.engine_class(name=engine_name, **self.config)
                new_engine.setup()
                self.engines[namespace] = new_engine
                self.export_metadata()
            msg = self.engines[namespace].handle_msg(msg)
        else:
            if namespace not in self.engines:
                return msg
            else:
                relevant_engines = [self.engines[namespace]]
            for engine in relevant_engines:
                msg = engine.handle_msg(msg)
        return msg

    def size(self) -> int:
        return sum(engine.size() for engine in self.engines.values())

    def get_stats(self):
        return {'size': self.size(), **{namespace: self.engines[namespace].size() for namespace in self.engines}}




#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import List
from loguru import logger
from pinecone.api_base import BaseAPI


__all__ = ["ControllerAPI"]


class ControllerAPI(BaseAPI):
    """API calls to the controller."""

    def get_service(self, service_name: str) -> str:
        """Returns service specs."""
        return self.get("/services/{}".format(service_name))["service"]

    def list_services(self) -> List[str]:
        """Returns the names of all of the services."""
        return self.get("/services")

    def get_status(self, service_name: str) -> dict:
        """Returns service status"""
        # TODO: service status should be an enum
        return self.get("/services/{}/status".format(service_name))

    def stop(self, service_name: str):
        """Stops an service."""
        response = self.delete("/services/{}".format(service_name))
        if not response["success"]:
            logger.error("Failed to stop the service '{}'. It probably wasn't running!".format(service_name))
        return response

    def deploy(self, service_json: str):
        """Deploys an service."""
        response = self.post("/services", json={'service': service_json})
        if response["success"]:
            logger.success("Successfully deployed {}".format(self.host))
        else:
            if "already exists" in response["msg"]:
                raise RuntimeError("Service already exists.")
            else:
                raise RuntimeError("Failed to deploy: {}".format(response["msg"]))
        return response



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import logging
import distutils.util
import os


def init_tracer():
    enable_tracing = distutils.util.strtobool(os.getenv('DD_TRACING_ENABLED', default='False'))
    if not enable_tracing:
        os.environ['DD_TRACE_STARTUP_LOGS'] = 'False'

    from ddtrace import tracer, patch_all
    from ddtrace.contrib import asyncio
    tracer.enabled = enable_tracing

    if enable_tracing:
        tracer.context_provider = asyncio.context_provider
        patch_all()


def init_profiler():
    enable_profiling = distutils.util.strtobool(os.getenv('DD_PROFILING_ENABLED', default='False'))
    if enable_profiling:
        import ddtrace.profiling.auto


def get_msg_size(msg: 'core_pb2.Request'):
    request_type = msg.WhichOneof('body')
    if request_type == 'info':
        return 1
    if request_type == "query":
        return msg.query.data.shape[0]
    if request_type == "index":
        return len(msg.index.ids)
    if request_type == "delete":
        return len(msg.delete.ids)
    if request_type == "fetch":
        return len(msg.fetch.ids)


def set_span_tags(span: 'Span', msg: 'core_pb2.Request'):
    span.set_tags({'path': msg.path,
                   'request_type': msg.WhichOneof('body'),
                   'service_name': msg.service_name})
    span.set_metric('msg_size', get_msg_size(msg))  # indicates the batch size


def set_span_service(span: 'Span', service_name: str):
    span.set_tags({'service_name': service_name})




from typing import Callable
from tqdm.auto import tqdm
import time
import os


class ProgressBar:
    DISABLE_PROGRESS_BAR = bool(os.getenv("PINECONE_DISABLE_PROGRESS_BAR"))

    def __init__(
        self,
        total: int,
        get_current_fn: Callable = None,
        get_remaining_fn: Callable = None,
        interval: float = 2,
        timeout: int = 3600,
        disable: bool = False,
        **kwargs
    ):
        """Progess Bar

        :param total: the total number of iterations
        :type total: int
        :param get_current_fn: get the current progress
        :type get_current_fn: Callable
        :param get_remaining_fn: get the remaining number of iterations
        :type get_remaining_fn: Callable
        :param interval: wait time between progress updates
        :type interval: float
        :param timeout: how long do we wait. Set to ``0`` to never time out.
        :type timeout: int
        :param disable: whether to disable the progress bar. Defaults to ``False``
        :type disable: bool
        :param kwargs: key word arguments tqdm
        :type kwargs: dict
        """
        self.total = total
        self.get_current_fn = get_current_fn
        self.get_remaining_fn = get_remaining_fn
        self.interval = interval
        self.timeout = timeout
        self.disable = disable or self.DISABLE_PROGRESS_BAR
        self.kwargs = kwargs
        self.progress = 0

    def update_progress(self):
        progress = None
        if self.get_current_fn:
            progress = self.get_current_fn()
        elif self.get_remaining_fn:
            progress = self.total - self.get_remaining_fn()
        else:
            raise RuntimeError("Unable to get current progress.")
        # NOTE: round float
        self.progress = round(progress, 4)

    def watch(self):
        tick = time.perf_counter()
        with tqdm(total=self.total, initial=self.progress, disable=self.disable, **self.kwargs) as pbar:
            while self.progress < self.total:
                time.sleep(self.interval)
                tock = time.perf_counter()
                if 0 < self.timeout < tock - tick:
                    raise RuntimeError("Time out when waiting for updates.")
                # Update progress only if it has changed.
                prev_progress = self.progress
                self.update_progress()
                pbar.update(round(self.progress - prev_progress, 4))

    @classmethod
    def iter(cls, iterable=None, **kwargs):
        """Returns vanilla tqdm."""
        disable = kwargs.pop("disable", cls.DISABLE_PROGRESS_BAR)
        return tqdm(iterable, disable=disable, **kwargs)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#
from typing import Iterable, List, Union, NamedTuple, Iterator, Callable, Any
import grpc, grpc._channel
import uuid
import atexit
import numpy as np

from pinecone import utils
from pinecone.utils.constants import MAX_MSG_SIZE
from pinecone.utils.progressbar import ProgressBar
from pinecone.constants import CLIENT_VERSION
from pinecone.protos import core_pb2_grpc, core_pb2

__all__ = ["GRPCClient"]


class GRPCClientConfig(NamedTuple):
    secure: bool = True
    timeout: int = 2
    conn_timeout: int = 1
    traceroute: bool = False
    reuse_channel: bool = True

    @classmethod
    def _from_dict(cls, kwargs: dict):
        cls_kwargs = {kk: vv for kk, vv in kwargs.items() if kk in cls._fields}
        return cls(**cls_kwargs)


class GRPCClient:
    """Client for gRPC."""

    def __init__(self, host: str = "", port: int = -1, service_name: str = "", api_key: str = "", **kwargs):
        self.host = host
        self.port = port
        self.metadata = (("api-key", api_key), ("service-name", service_name))
        config = GRPCClientConfig._from_dict(kwargs)
        self.secure = config.secure
        self.timeout = config.timeout
        self.traceroute = config.traceroute
        self.conn_timeout = config.conn_timeout
        self.reuse_channel = config.reuse_channel
        self._channel = None
        atexit.register(_clean_up_client, client=self)

    @property
    def channel(self):
        """Creates GRPC channel."""

        def _gen():
            target = "{}:{}".format(self.host, self.port)
            options = (
                ("grpc.max_send_message_length", MAX_MSG_SIZE),
                ("grpc.max_receive_message_length", MAX_MSG_SIZE),
            )
            if not self.secure:
                return grpc.insecure_channel(target, options=options)
            tls = grpc.ssl_channel_credentials()
            return grpc.secure_channel(target, tls, options=(("grpc.ssl_target_name_override", self.host),) + options)

        if self.reuse_channel and self._channel and self.grpc_server_on():
            return self._channel
        self._channel = _gen()
        return self._channel

    def get_index_request(
        self,
        ids: List[str] = None,
        data: Union[np.ndarray, List] = None,
        path: str = None,
        namespace: str = None,
    ) -> "core_pb2.Request":
        """Returns an upsert request."""
        req = core_pb2.IndexRequest()
        if ids is not None:
            req.ids[:] = [str(ee) for ee in ids]
        if data is not None:
            req.data.CopyFrom(utils.dump_numpy(_to_ndarray(data)))
        return core_pb2.Request(
            request_id=_generate_request_id(),
            version=CLIENT_VERSION,
            index=req,
            path=path,
            namespace=namespace,
            timeout=self.timeout,
            traceroute=self.traceroute,
        )

    def get_delete_request(
        self,
        ids: Iterable[str],
        path: str = None,
        namespace: str = None,
    ) -> "core_pb2.Request":
        """Returns a delete request."""
        req = core_pb2.DeleteRequest(ids=ids)
        return core_pb2.Request(
            request_id=_generate_request_id(),
            version=CLIENT_VERSION,
            delete=req,
            path=path,
            namespace=namespace,
            timeout=self.timeout,
            traceroute=self.traceroute,
        )

    def get_fetch_request(
        self,
        ids: Iterable[str],
        path: str = None,
        namespace: str = None,
    ) -> "core_pb2.Request":
        """Returns a fetch request."""
        req = core_pb2.FetchRequest(ids=ids)
        return core_pb2.Request(
            request_id=_generate_request_id(),
            version=CLIENT_VERSION,
            fetch=req,
            path=path,
            namespace=namespace,
            timeout=self.timeout,
            traceroute=self.traceroute,
        )

    def get_query_request(
        self,
        top_k: int = None,
        include_data: bool = None,
        data: Union[np.ndarray, List] = None,
        matches: List[dict] = None,
        path: str = None,
        namespace: str = None,
    ) -> "core_pb2.Request":
        """Returns a query request."""
        req = core_pb2.QueryRequest()
        if top_k is not None:
            req.top_k = top_k
        if include_data is not None:
            req.include_data = include_data
        if data is not None:
            req.data.CopyFrom(utils.dump_numpy(_to_ndarray(data)))
        if matches is not None:
            req.matches.extend(
                [
                    core_pb2.ScoredResults(
                        ids=mat.get("ids"),
                        scores=mat.get("scores"),
                        data=utils.dump_numpy(np.array(mat["data"])) if mat.get("data") else None,
                    )
                    for mat in matches
                ]
            )
        return core_pb2.Request(
            request_id=_generate_request_id(),
            version=CLIENT_VERSION,
            query=req,
            path=path,
            namespace=namespace,
            timeout=self.timeout,
            traceroute=self.traceroute,
        )

    def get_info_request(
        self,
        path: str = None,
        namespace: str = None,
    ) -> "core_pb2.Request":
        """Returns an info request"""
        return core_pb2.Request(
            request_id=_generate_request_id(),
            version=CLIENT_VERSION,
            info=core_pb2.InfoRequest(),
            path=path,
            namespace=namespace,
            timeout=self.timeout,
            traceroute=self.traceroute,
        )

    def grpc_server_on(self) -> bool:
        try:
            grpc.channel_ready_future(self._channel).result(timeout=self.conn_timeout)
            return True
        except grpc.FutureTimeoutError:
            return False

    def stream_requests(self, request_stream):
        """Creates sub to send proptobuf requests."""
        stub = core_pb2_grpc.RPCClientStub(self.channel)
        if self.conn_timeout > 0 and not self.grpc_server_on():
            raise ConnectionError("Failed to connect to gRPC host {}:{}".format(self.host, self.port))
        exception = None
        try:
            for response in stub.Call(request_stream, metadata=self.metadata):
                yield response
        except grpc._channel._MultiThreadedRendezvous as e:
            exception = e
        if exception:
            raise ConnectionError(
                "Failed to connect to gRPC remote {}:{}. Status = {}. Details = {}. Debug = {}".format(
                    self.host, self.port, exception._state.code, exception.details(), exception.debug_error_string()
                )
            )

    def send_request(self, request):
        stub = core_pb2_grpc.RPCClientStub(self.channel)
        if self.conn_timeout > 0 and not self.grpc_server_on():
            raise ConnectionError("Failed to connect to gRPC host {}:{}".format(self.host, self.port))
        try:
            return stub.CallUnary(request, metadata=self.metadata)
        except grpc._channel._MultiThreadedRendezvous as e:
            exception = e
        raise ConnectionError(
            "Failed to connect to gRPC remote {}:{}. Status = {}. Details = {}. Debug = {}".format(
                self.host, self.port, exception._state.code, exception.details(), exception.debug_error_string()
            )
        )


class Cursor:
    _DEFAULT_BATCH_SIZE = 100

    def __init__(
        self,
        items: Iterable[np.ndarray],
        grpc_client: GRPCClient,
        create_request_fn: Callable[[Iterable], "core_pb2.Request"],
        parse_response_fn: Callable[["core_pb2.Request"], Iterable] = None,
        unary: bool = False,
        batch_size: int = None,
        disable_progress_bar: bool = False,
    ):
        """A cursor batch-processes requests sent to a service.

        :param config: cursor configurations
        :type config: :class:`CursorConfig`
        :param items: Items to process
        :type items: Iterable[np.ndarray]
        :param batch_size: the number of items to batch-process, defaults to 100
        :type batch_size: int, optional
        """
        self.batch_size = batch_size or self._DEFAULT_BATCH_SIZE
        self._item_iterator = iter(items)
        self.unary = unary
        self.disable_progress_bar = disable_progress_bar
        self.client = grpc_client
        self.create_reqeust_fn = create_request_fn
        self.parse_response_fn = parse_response_fn or self.identity

    def collect(self) -> Iterable:
        """Processes all of the items and returns the results."""
        result = []
        if self.disable_progress_bar:
            result = list(self._fetch_unbuffered())
        else:
            result = list(ProgressBar.iter(self._fetch_unbuffered()))
        return result

    def stream(self) -> Iterator:
        """Processes all of the items, returns the results one batch at a time."""
        ret_iter = None
        if self.disable_progress_bar:
            ret_iter = self._fetch_unbuffered()
        else:
            ret_iter = ProgressBar.iter(self._fetch_unbuffered())
        return ret_iter

    def __iter__(self):
        raise RuntimeError(
            "To iterate through the results, " "consider using one of the followoing methods: `collect()`, `stream()`."
        )

    @classmethod
    def identity(cls, entity: Any):
        """An identity function."""
        return entity

    def _fetch_unbuffered(self) -> Iterator:
        """Creates a generator to fetch results."""
        if self.unary:
            for request in self._generate_requests():
                response_ = self.client.send_request(request)
                for elem in self._parse_response(response_):
                    yield elem
        else:
            for response_ in self.client.stream_requests(self._generate_requests()):
                for elem in self._parse_response(response_):
                    yield elem

    def _iter_by_batch(self) -> Iterable[np.ndarray]:
        """Iterate the items in batches."""
        batch = []
        for item in self._item_iterator:
            batch.append(item)
            if len(batch) >= self.batch_size:
                yield batch
                batch = []
        if batch:
            yield batch

    def _generate_requests(self) -> Iterator["core_pb2.Request"]:
        """Generates protobuf requests in batches."""
        for batch in self._iter_by_batch():
            yield self.create_reqeust_fn(batch)

    def _parse_response(self, response: "core_pb2.Request") -> Iterable:
        if response.status.code == core_pb2.Status.StatusCode.Value("ERROR"):
            exception = bool(response.status.details) and response.status.details[0].exception
            error_msg = exception or response.status.description or "Unknown error when fetching results."
            raise RuntimeError(error_msg)

        return self.parse_response_fn(response)


def _to_ndarray(arr: Iterable) -> np.ndarray:
    """Convert an array to a numpy array by making best guesses."""
    if isinstance(arr, np.ndarray):
        return arr
    else:
        return np.asarray(arr)


def _generate_request_id() -> int:
    return uuid.uuid4().int & (1 << 64) - 1


def _clean_up_client(client: GRPCClient):
    """Cleans up client resources"""
    try:
        client.channel.close()
    except TypeError:
        pass



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#
from typing import NamedTuple
import os
import sentry_sdk
import configparser
from pinecone.utils import get_version, get_environment

__all__ = ["CLIENT_VERSION", "Config", "PACKAGE_ENVIRONMENT", "SENTRY_DSN_TXT_RECORD"]

PACKAGE_VERSION = get_version()
PACKAGE_ENVIRONMENT = get_environment() or "development"
CLIENT_VERSION = "0.1"
SENTRY_DSN_TXT_RECORD = "pinecone-client.sentry.pinecone.io"


def _set_sentry_tags(config: dict):
    sentry_sdk.set_tag("package_version", PACKAGE_VERSION)
    for key, val in config.items():
        sentry_sdk.set_tag(key, val)


class ConfigBase(NamedTuple):
    environment: str = ""
    api_key: str = ""
    controller_host: str = ""
    hub_host: str = ""
    hub_registry: str = ""
    base_image: str = ""


class _CONFIG:
    """

    Order of configs to load:

    - configs specified explictly in reset
    - environment variables
    - configs specified in the INI file
    - default configs
    """

    def __init__(self):
        self.reset()

    def reset(self, config_file=None, **kwargs):
        config = ConfigBase()

        # Load config from file
        file_config = self._load_config_file(config_file)

        # Get the environment first. Make sure that it is not overwritten in subsequent config objects.
        environment = (
            kwargs.pop("environment", None)
            or os.getenv("PINECONE_ENVIRONMENT")
            or file_config.pop("environment", None)
            or "beta"
        )
        config = config._replace(environment=environment)

        # Set default config
        default_config = ConfigBase(
            controller_host="https://controller.{0}.pinecone.io".format(config.environment),
            hub_host="https://hub-api.{0}.pinecone.io".format(config.environment),
            hub_registry="https://hub.{0}.pinecone.io".format(config.environment),
            base_image="hub.{0}.pinecone.io/pinecone/base:{1}".format(config.environment, PACKAGE_VERSION),
        )
        config = config._replace(**self._filter_config(default_config._asdict()))

        # Set INI file config
        config = config._replace(**self._filter_config(file_config))

        # Set environment config
        env_config = ConfigBase(
            api_key=os.getenv("PINECONE_API_KEY"),
            controller_host=os.getenv("PINECONE_CONTROLLER_HOST"),
            hub_host=os.getenv("PINECONE_HUB_HOST"),
            hub_registry=os.getenv("PINECONE_HUB_REGISTRY"),
            base_image=os.getenv("PINECONE_BASE_IMAGE"),
        )
        config = config._replace(**self._filter_config(env_config._asdict()))

        # Set explicit config
        config = config._replace(**self._filter_config(kwargs))

        self._config = config

        # Sentry
        _set_sentry_tags(self._config._asdict())

    def _filter_config(self, config: dict) -> dict:
        """Retain only the valid keys and non-null values.

        Also removes the "environment" config from the config object.
        """
        return {
            key: val
            for key, val in config.items()
            if key != "environment" and key in ConfigBase._fields and val is not None
        }

    def _load_config_file(self, config_file: str) -> dict:
        """Load from INI config file."""
        config_obj = {}
        if config_file:
            full_path = os.path.expanduser(config_file)
            if os.path.isfile(full_path):
                parser = configparser.ConfigParser()
                parser.read(full_path)
                if "default" in parser.sections():
                    config_obj = {**parser["default"]}
        return config_obj

    @property
    def ENVIRONMENT(self):
        return self._config.environment

    @property
    def API_KEY(self):
        return self._config.api_key

    @property
    def CONTROLLER_HOST(self):
        return self._config.controller_host

    @property
    def HUB_HOST(self):
        return self._config.hub_host

    @property
    def HUB_REGISTRY(self):
        return self._config.hub_registry

    @property
    def BASE_IMAGE(self):
        return self._config.base_image


Config = _CONFIG()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import os
import enum
from pinecone.utils import get_container_memory_limit

RECV_TIMEOUT = 0.5
SEND_TIMEOUT = 0.5

MAX_CLIENTS = 100
MAX_RETRY_MSG = 100
DEFAULT_TIMEOUT = 2
GATEWAY_NAME = 'gatewayrouter'
AGGREGATOR_NAME = 'aggregator'
MAX_MSG_SIZE = 64 * 1024 * 1024
MAX_MSGS_PER_CONNECTION = 100
MAX_SOCKS_OPEN = 1000
DNS_TTL = 5

ZMQ_CONTROL_PORT = os.environ.get('ZMQ_CONTROL_PORT', 5559)
ZMQ_PORT_IN = os.environ.get('ZMQ_PORT_IN', 5557)
ZMQ_LOG_PORT = os.environ.get('ZMQ_LOG_PORT', 5558)
PC_CONTROLLER_PORT = os.environ.get('PC_CONTROLLER_PORT', 8083)
SERVICE_GATEWAY_PORT = os.environ.get('SERVICE_GATEWAY_PORT', 5007)
PERSISTENT_VOLUME_MOUNT = '/data'
ENV_VARS = [ZMQ_CONTROL_PORT, ZMQ_PORT_IN, ZMQ_LOG_PORT, PC_CONTROLLER_PORT]
ZMQ_SECONDARY_PORT = 5559
MAX_MSGS_PENDING = 1000

MAX_ID_LENGTH = int(os.environ.get("PINECONE_MAX_ID_LENGTH", default="64"))
MEMORY_UTILIZATION_LIMIT = float(os.environ.get("MEMORY_UTILIZATION_LIMIT", default="0.9"))
MEMORY_LIMIT_BYTES = MEMORY_UTILIZATION_LIMIT * get_container_memory_limit()


CPU_UNIT = 500
MEMORY_UNIT = 1875
DISK_UNIT = 5


class NodeType(str, enum.Enum):
    STANDARD = 'STANDARD'
    COMPUTE = 'COMPUTE'
    MEMORY = 'MEMORY'
    STANDARD2X = 'STANDARD2X'
    COMPUTE2X = 'COMPUTE2X'
    MEMORY2X = 'MEMORY2X'
    STANDARD4X = 'STANDARD4X'
    COMPUTE4X = 'COMPUTE4X'
    MEMORY4X = 'MEMORY4X'


NODE_TYPE_RESOURCES = {NodeType.STANDARD: {'memory': MEMORY_UNIT, 'cpu': CPU_UNIT, 'disk': DISK_UNIT},
                       NodeType.STANDARD2X: {'memory': 2 * MEMORY_UNIT, 'cpu': 2 * CPU_UNIT, 'disk': 2 * DISK_UNIT},
                       NodeType.STANDARD4X: {'memory': 4 * MEMORY_UNIT, 'cpu': 4 * CPU_UNIT, 'disk': 4 * DISK_UNIT},
                       NodeType.COMPUTE: {'memory': MEMORY_UNIT, 'cpu': 2 * CPU_UNIT, 'disk': DISK_UNIT},
                       NodeType.COMPUTE2X: {'memory': 2 * MEMORY_UNIT, 'cpu': 4 * CPU_UNIT, 'disk': DISK_UNIT * 2},
                       NodeType.COMPUTE4X: {'memory': 4 * MEMORY_UNIT, 'cpu': 8 * CPU_UNIT, 'disk': DISK_UNIT * 4},
                       NodeType.MEMORY: {'memory': 2 * MEMORY_UNIT, 'cpu': CPU_UNIT, 'disk': DISK_UNIT * 2},
                       NodeType.MEMORY2X: {'memory': 4 * MEMORY_UNIT, 'cpu': 2 * CPU_UNIT, 'disk': DISK_UNIT * 4},
                       NodeType.MEMORY4X: {'memory': 8 * MEMORY_UNIT, 'cpu': 4 * CPU_UNIT, 'disk': DISK_UNIT * 8}}




#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import NamedTuple
from pinecone import __version__
from pinecone.api_base import BaseAPI

__all__ = ["ActionAPI"]


class WhoAmIResponse(NamedTuple):
    username: str


class VersionResponse(NamedTuple):
    server: str
    client: str


class ActionAPI(BaseAPI):
    """User related API calls."""

    def whoami(self) -> WhoAmIResponse:
        """Returns user information."""
        response = self.get("/actions/whoami")
        return WhoAmIResponse(username=response.get("user_name", "UNDEFINED"))

    def version(self) -> VersionResponse:
        """Returns version information."""
        response = self.get("/actions/version")
        return VersionResponse(server=response.get("version", "UNKNOWN"),
                               client=__version__)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.network.selector import SocketSelector

from pinecone.utils import load_numpy, dump_numpy
from pinecone.protos import core_pb2
from typing import List, Optional
import hashlib


def hash_id(item_id: str) -> int:
    return int(hashlib.md5(item_id.encode('utf-8')).hexdigest(), 16)


class ShardSelector(SocketSelector):

    @classmethod
    def shard_msg(cls, msg: 'core_pb2.Request', req_type: str, shards: int) -> List[Optional['core_pb2.Request']]:
        req = getattr(msg, req_type)
        hashed_ids = [hash_id(id_) for id_ in req.ids]
        shard_msgs = [core_pb2.Request() for i in range(shards)]
        data = load_numpy(req.data) if req_type == 'index' else None
        for shard_id, new_msg in enumerate(shard_msgs):
            new_msg.CopyFrom(msg)
            relevant_indices = [i for i in range(len(hashed_ids)) if (
                    hashed_ids[i] % shards) == shard_id]
            new_req = getattr(new_msg, req_type)
            relevant_ids = [new_req.ids[i] for i in relevant_indices]
            del new_req.ids[:]
            new_req.ids.extend(relevant_ids)
            if data is not None and len(data) > 0:
                new_req.data.CopyFrom(dump_numpy(data[relevant_indices, :]))
        return shard_msgs

    @classmethod
    def select_socket(cls, msg: 'core_pb2.Request', shards: int) -> List[Optional['core_pb2.Request']]:
        req_type = msg.WhichOneof('body')
        if req_type in ['info', 'query', 'fetch'] or shards <= 1:
            return [msg]
        return ShardSelector.shard_msg(msg, req_type, shards)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from collections import defaultdict
import graphviz
import networkx as nx

import pinecone
from pinecone.specs import service as service_specs
from pinecone.functions import Function
from pinecone.functions.model import HubFunction
from pinecone.functions.engine import namespaced
from pinecone.functions.ranker import aggregator
from pinecone.utils.sentry import sentry_decorator as sentry
from pinecone.utils.constants import NodeType
from typing import Dict, NamedTuple, Optional

__all__ = ["Graph", "IndexGraph"]


class Graph(service_specs.Service):
    """The graphical representation of a service."""

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

    @sentry
    def dump(self, format: str = "yaml") -> str:
        """Dumps the graph as yaml or json string.

        :param format: one of {"yaml", "json"}, defaults to "yaml".
        :type format: str, optional
        :rtype: str
        """
        """Dumps current graph as yaml or json."""
        if format == "yaml":
            return self.to_yaml()
        elif format == "json":
            return self.to_json()
        else:
            raise NotImplementedError("Format {format} not supported.".format(format=format))

    @sentry
    def to_graph(self) -> nx.MultiDiGraph:
        """Exports to a `networkx <https://networkx.org/>`_ graph.

        Construct a multi directed graph.
        Add incoming traffic and outgoing traffic nodes to signify the traffic flow.

        :return: [description]
        :rtype: nx.MultiDiGraph
        """

        def _get_node_name(fn: "pinecone.function.Function"):
            image_name = fn.image if isinstance(fn, HubFunction) else ""
            if image_name:
                return '"{}\n<{}>"'.format(fn.name, image_name)
            else:
                return fn.name

        def _get_edge_name(path_name: str):
            return "[{}]".format(path_name)

        graph = nx.MultiDiGraph()

        for path_name, steps in self.paths.items():
            node_names = [_get_node_name(self.functions[ss]) for ss in steps]
            for start_node, end_node in zip(["incoming traffic", *node_names], [*node_names, "outgoing traffic"]):
                graph.add_edge(
                    start_node,
                    end_node,
                    label=_get_edge_name(path_name),
                    key=_get_edge_name(path_name),
                )
        return graph

    @sentry
    def view(self):
        """Visualizes the graph in an iPython notebook.

        .. note::
            This method requires the `graphviz <https://graphviz.org/download/>`_ package
            installed on your operating system.
        """
        multigraph = self.to_graph()
        union_labels = defaultdict(list)
        for start_node, end_node, label in multigraph.edges.data("label"):
            if label:
                union_labels[(start_node, end_node)].append(label)

        graph = nx.DiGraph()
        for start_node, end_node, path in multigraph.edges:
            labels = union_labels.get((start_node, end_node)) or []
            graph.add_edge(start_node, end_node, label="".join(labels))

        return graphviz.Source(nx.nx_pydot.to_pydot(graph))


class IndexConfig(NamedTuple):
    """Index configuration options.

    :param engine_type: type of engine, one of {"approximated", "exact"}, defaults to "approximated".
        The "approximated" engine uses fast approximate search algorithms developed by Pinecone.
        The "exact" engine uses accurate exact search algorithms.
        It performs exhaustive searches and thus it is usually slower than the "approximated" engine.
    :type engine_type: str, optional
    :param metric: type of metric used in the vector index, one of {"cosine", "dotproduct", "euclidean"}, defaults to "cosine".
        Use "cosine" for cosine similarity,
        "dotproduct" for dot-product,
        and "euclidean" for euclidean distance.
    :type metric: str, optional
    :param shards: the number of shards for the engine, defaults to 1.
        As a general guideline, use 1 shard per 1 GB of data.
    :type shards: int, optional
    :param replicas: the number of replicas, defaults to 1.
        Use at least 2 replicas if you need high availability (99.99% uptime) for querying.
        For additional throughput (QPS) your service needs to support, provision additional replicas.
    :type replicas: int, optional
    :param gateway_replicas: number of replicas of both the gateway and the aggregator.
    :type gateway_replicas: int
    :param engine_args: advanced arguments for the engine instance in the graph.
    :type engine_args: dict
    """
    engine_type: str = "approximated"
    metric: str = "cosine"
    shards: int = 1
    replicas: int = 1
    gateway_replicas: int = 1
    node_type: str = NodeType.STANDARD.value
    engine_args: Optional[dict] = None

    @classmethod
    def _from_dict(cls, kwargs: dict):
        cls_kwargs = {kk: vv for kk, vv in kwargs.items() if kk in cls._fields}
        return cls(**cls_kwargs)

    @classmethod
    def _from_graph(cls, graph: Graph):
        """Reconstruct configs from a Graph."""
        config = {}
        config["gateway_replicas"] = graph.gateway_replicas
        engine = graph.functions.get("engine")
        if engine:
            config["replicas"] = engine.replicas
            config["shards"] = engine.shards
            config["node_type"] = engine._node_type.value
            config["engine_type"] = engine.config.pop("engine_type")
            config["metric"] = engine.config.pop("metric")
            config["engine_args"] = {**(engine.config or {})}
        return cls(**config)


class IndexGraph(Graph):
    """The graphical representation of an index service.

    An index service consists of *preprocessors*, the *vector index*, and *postprocessors*.
    Preprocessors modify the items before indexing them, or modify the queries before retrieving results
    from the index.
    Examples of preprocessors include matrix fatorization models for transforming movie ids into embeddings,
    and BERT transformers for text data.
    Postprocessors modify the retrieved results of queries. Examples of postprocessors include
    re-rankers for balancing recommendation fairness, or imputation functions for increasing diversity.
    """

    _default_engine = namespaced.NamespacedEngine
    _default_aggregator = aggregator.Aggregator

    @sentry
    def __init__(self, **kwargs):
        """See :class:`IndexConfig` for details about the parameters."""
        config = IndexConfig._from_dict(kwargs)
        engine_type = config.engine_type
        metric = config.metric
        shards = config.shards
        replicas = config.replicas
        gateway_replicas = config.gateway_replicas
        node_type = NodeType(config.node_type)
        engine_args = config.engine_args or {}

        super().__init__(gateway_replicas=gateway_replicas)
        self.preprocessors = {
            "read": [],
            "write": [],
        }
        self.postprocessors = []

        # The engine and the aggregator must be specified for an index service
        self.engine = self._default_engine(
            engine_type,
            metric=metric,
            replicas=replicas,
            shards=shards,
            name="engine",
            node_type=node_type,
            **engine_args
        )
        self.aggregator = (
            None
            if shards == 1
            else self._default_aggregator(num_shards=shards, name="aggregator", replicas=gateway_replicas)
        )
        self._update_graph()

    @sentry
    def add_postprocessor(self, image_uri: str = None, config: dict = None, fn: Function = None):
        """Adds a postprocessor to the graph.

        You can add mutiple postprocessors to a graph.
        Postprocessors are on the "read" path of a graph
        and modify results of a ``query`` request.

        :param image_uri: a Pinecone Model Hub docker image URI, defaults to None
        :type image_uri: str, optional
        :param config: configurations for :class:`HubFunction`, defaults to None
        :type config: dict, optional
        :param fn: an instance of a Pinecone Function, defaults to None.
            For custom functions, use :class:`HubFunction`.
            If specified, ``image_uri`` and ``config`` will be ignored.
        :type fn: :class:`Function`, optional
        """
        _fn = fn or HubFunction(image=image_uri, config=config)
        self.postprocessors.append(_fn)
        self._update_graph()

    @sentry
    def remove_postprocessor(self, name: str):
        """Removes a postprocessor."""
        self.postprocessors = [fn for fn in self.postprocessors if fn.name != name]
        self._update_graph()

    @sentry
    def add_preprocessor(self, path: str, image_uri: str = None, config: dict = None, fn: Function = None):
        """Adds a preprocessor to the given path of the graph.

        You can add multiple preprocessors to the same path in a graph.
        ``query`` requests flow through the read path of the graph,
        and ``upsert`` requests flow through the write path of the graph.

        :param path: the path in the graph to add the preprocessor to
        :type path: {"read", "write"}
        :param image_uri: a Pinecone Model Hub docker image URI, defaults to None
        :type image_uri: str, optional
        :param config: configurations for :class:`HubFunction`, defaults to None
        :type config: dict, optional
        :param fn: an instance of a Pinecone Function, defaults to None.
            For custom functions, use :class:`HubFunction`.
            If specified, ``image_uri`` and ``config`` will be ignored.
        :type fn: :class:`Function`, optional
        """
        if path not in {"read", "write"}:
            raise RuntimeError("'path' must be one of {0}".format({"read", "write"}))
        _fn = fn or HubFunction(image=image_uri, config=config)
        self.preprocessors[path].append(_fn)
        self._update_graph()

    @sentry
    def remove_preprocessor(self, path: str, name: str):
        """Removes a preprocessor from the given path.

        :param path: the path in the graph to remove the preprocessor from
        :type path: {"read", "write"}
        :param name: name of the preprocessor to remove
        :type name: str
        """
        if path not in {"read", "write"}:
            raise RuntimeError("'path' must be one of {0}".format({"read", "write"}))
        self.preprocessors[path] = [fn for fn in self.preprocessors[path] if fn.name != name]
        self._update_graph()

    @sentry
    def add_read_preprocessor(self, image_uri: str = None, config: dict = None, fn: Function = None):
        """Adds a preprocessor to the "read" path of the graph.

        See :func:`add_preprocessor` for details.
        """
        self.add_preprocessor(path="read", image_uri=image_uri, config=config, fn=fn)

    @sentry
    def remove_read_preprocessor(self, name: str):
        """Removes a preprocessor from the "read" path.

        See :func:`remove_preprocessor` for details.

        """
        self.remove_preprocessor(path="read", name=name)

    @sentry
    def add_write_preprocessor(self, image_uri: str = None, config: dict = None, fn: Function = None):
        """Adds a preprocessor to the "write" path of the graph.

        See :func:`add_preprocessor` for details.
        """
        self.add_preprocessor(path="write", image_uri=image_uri, config=config, fn=fn)

    @sentry
    def remove_write_preprocessor(self, name: str):
        """Removes a preprocessor from the "write" path.

        See :func:`remove_preprocessor` for details.

        """
        self.remove_preprocessor(path="write", name=name)

    def _update_graph(self):
        write_functions = [*self.preprocessors["write"], self.engine, self.aggregator]
        read_functions = [*self.preprocessors["read"], self.engine, self.aggregator, *self.postprocessors]
        self.paths["write"] = [fn.name for fn in write_functions if fn]
        self.paths["read"] = [fn.name for fn in read_functions if fn]
        self.functions = {fn.name: fn for fn in [*write_functions, *read_functions] if fn}



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Optional, Iterable, List

import numpy as np

from pinecone.functions.ranker import Ranker


class Aggregator(Ranker):

    def score(self, q: np.ndarray, vectors: np.ndarray, prev_scores: Iterable[float]) -> List[float]:
        return list(prev_scores)

    @property
    def volume_request(self) -> Optional[int]:
        return 0



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import List
from loguru import logger
from pinecone.api_base import BaseAPI

__all__ = ["RouterAPI"]


class RouterAPI(BaseAPI):
    """API calls to the controller."""

    def get_router(self, router_name):
        """Returns the spec of a router."""
        return self.get("/routers/{}".format(router_name))['router']

    def deploy(self, router_json: str):
        """Creates or updates a traffic router."""
        res_json = self.post("/routers", json={"router": router_json})
        if res_json["success"]:
            logger.success("Successfully deployed to {}".format(self.host))
        else:
            # TODO: improve error handling
            if "already exists" in res_json["msg"]:
                raise RuntimeError("Service already exists.")
            else:
                raise RuntimeError("Failed to deploy: {}".format(res_json["msg"]))
        return res_json

    def list(self) -> List[str]:
        """Returns the names of all traffic routers."""
        return self.get("/routers")

    def stop(self, router_name: str):
        """Stops a traffic router."""
        response = self.delete("/routers/{}".format(router_name))
        if not response["success"]:
            logger.error("Failed to stop {0}, it probably wasn't running!".format(router_name))
        return response

    def get_status(self, router_name: str) -> dict:
        """Gets status of a traffic router."""
        return self.get("/routers/{}/status".format(router_name))



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import List, NamedTuple, Tuple
import time

from loguru import logger

from .api_controller import ControllerAPI
from .constants import Config
from .graph import Graph
from pinecone.utils.sentry import sentry_decorator as sentry
from pinecone.utils.progressbar import ProgressBar

__all__ = ["describe", "deploy", "stop", "ls", "ServiceMeta"]


class ServiceMeta(NamedTuple):
    """Metadata of a service."""

    name: str
    graph: Graph
    status: dict


def _get_controller_api():
    return ControllerAPI(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)


@sentry
def describe(service_name: str) -> ServiceMeta:
    """Returns the metadata of a service.

    :param service_name: name of the service
    :type service_name: str
    :return: :class:`ServiceMeta`
    """
    api = _get_controller_api()
    service_json = api.get_service(service_name)
    graph = Graph.from_json(service_json) if service_json else None
    return ServiceMeta(name=graph.name, graph=graph, status=api.get_status(service_name) or {})


@sentry
def deploy(service_name: str, graph: Graph, wait: bool = True, **kwargs) -> Tuple[dict, ProgressBar]:
    """Create a new Pinecone service from the graph.

    :param service_name: name of the service
    :type service_name: str
    :param graph: the graphical representation fo the service
    :type graph: :class:`pinecone.graph.Graph`
    :param wait: wait for the service to deploy. Defaults to ``True``
    :type wait: bool
    """
    graph.validate()

    # copy graph
    graph_ = Graph.from_json(graph.to_json())
    graph_.name = service_name

    api = _get_controller_api()

    if service_name in api.list_services():
        raise RuntimeError(
            "A service with the name '{}' already exists. Please deploy your service with a different name.".format(
                service_name
            )
        )
    else:
        response = api.deploy(graph_.to_json())

    # Wait for service to deploy
    status = api.get_status(service_name)
    total_deployments = len(status.get("waiting") or []) + len(status.get("crashed") or [])

    def get_remaining():
        """Get the number of pods that still need to be deployed."""
        status = api.get_status(service_name)
        logger.info("Deployment status: waiting={}, crashed={}".format(status.get("waiting"), status.get("crashed")))
        remaining_deployments = len(status.get("waiting") or []) + len(status.get("crashed") or [])
        return remaining_deployments

    pbar = ProgressBar(total=total_deployments, get_remaining_fn=get_remaining)
    if wait:
        pbar.watch()
    return response, pbar


@sentry
def stop(service_name: str, wait: bool = True, **kwargs) -> Tuple[dict, ProgressBar]:
    """Stops a service.

    :param service_name: name of the service
    :type service_name: str
    :param wait: wait for the service to deploy. Defaults to ``True``
    :type wait: bool
    """
    api = _get_controller_api()
    response = api.stop(service_name)

    # Wait for service to stop
    def get_remaining():
        return 1 * (service_name in api.list_services())

    pbar = ProgressBar(total=1, get_remaining_fn=get_remaining)
    if wait:
        pbar.watch()
    return response, pbar


@sentry
def ls() -> List[str]:
    """Returns all services names."""
    api = _get_controller_api()
    return api.list_services()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import abc
from typing import List, Optional


class Runnable:

    @abc.abstractmethod
    def to_args(self):
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_args(cls, args):
        raise NotImplementedError

    @abc.abstractmethod
    def start(self):
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def image(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def replicas(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def shards(self) -> int:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def ports(self) -> List[int]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def ext_port(self) -> Optional[int]:
        raise NotImplementedError

    @property
    def memory_request(self) -> int:
        return 200

    @property
    def cpu_request(self) -> int:
        """In 1/1000th of cpu"""
        return 100

    @property
    def volume_request(self) -> Optional[int]:
        """
        Either the size of requested persistent volume in GB or None if ephemeral
        :return:
        """
        return None



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function
from pinecone.utils.constants import NODE_TYPE_RESOURCES
from abc import abstractmethod
from typing import Dict, Optional


class Engine(Function):

    """
    Generic interface for defining an engine
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    def get_stats(self) -> Dict:
        return {'size': self.size()}

    @property
    def stats_frequency_seconds(self):
        return 60

    @property
    def memory_request(self) -> int:
        return NODE_TYPE_RESOURCES[self._node_type]['memory']

    @property
    def cpu_request(self) -> int:
        return NODE_TYPE_RESOURCES[self._node_type]['cpu']

    @property
    def volume_request(self) -> Optional[int]:
        return NODE_TYPE_RESOURCES[self._node_type]['disk']



from typing import NamedTuple, Optional, Tuple, List
import enum
from loguru import logger

from pinecone.utils.sentry import sentry_decorator as sentry
from pinecone.utils.progressbar import ProgressBar
from .service import deploy as service_deploy, stop as service_stop, ls as service_ls, describe as service_describe
from .graph import IndexGraph, IndexConfig

__all__ = [
    "create",
    "delete",
    "ls",
    "describe",
    "create_index",
    "delete_index",
    "describe_index",
    "list_indexes",
    "ResourceType",
    "ResourceDescription",
]


class ResourceType(enum.Enum):
    INDEX = "index"


class ResourceDescription(NamedTuple):
    """Description of a resource."""

    name: str
    kind: str
    status: dict
    config: dict


@sentry
def create(name: str, kind: str = "index", wait: bool = True, **kwargs) -> Optional[Tuple[dict, ProgressBar]]:
    """Creates a Pinecone resource.

    :param name: the name of the resource.
    :type name: str
    :param kind: what kind of resource. Defaults to ``index``.
    :type kind: str
    :param wait: wait for the resource to deploy. Defaults to ``True``
    :type wait: bool
    :param `**kwargs`: see resource-specific configurations.
        For example, you can refer to :class:`IndexConfig` for
        configuration options for a Pinecone index.
    """
    if kind == ResourceType.INDEX.value:
        return service_deploy(service_name=name, graph=IndexGraph(**kwargs), wait=wait)
    logger.warning("Unrecognized resource type '{}'.".format(kind))


@sentry
def delete(name: str, kind: str = "index", wait: bool = True) -> Optional[Tuple[dict, ProgressBar]]:
    """Deletes a Pinecone resource.

    :param name: the name of the resource.
    :type name: str
    :param kind: what kind of resource. Defaults to ``index``.
    :type kind: str
    :param wait: wait for the resource to deploy. Defaults to ``True``
    :type wait: bool
    """
    if kind == ResourceType.INDEX.value:
        return service_stop(service_name=name, wait=wait)
    logger.warning("Unrecognized resource type '{}'.".format(kind))


@sentry
def ls(kind: str = "index") -> Optional[List[str]]:
    """Lists all resources of a certain kind.

    :param kind: what kind of resource. Defaults to ``index``.
    :type kind: str
    """
    if kind == ResourceType.INDEX.value:
        return service_ls()
    logger.warning("Unrecognized resource type '{}'.".format(kind))


@sentry
def describe(name: str, kind: str = "index") -> Optional[ResourceDescription]:
    """Describes the resource.

    :param name: the name of the resource.
    :type name: str
    :param kind: what kind of resource. Defaults to ``index``.
    :type kind: str
    """
    if kind == ResourceType.INDEX.value:
        desc = service_describe(service_name=name)
        graph = desc.graph
        config = IndexConfig._from_graph(graph)._asdict()
        return ResourceDescription(name=desc.name, kind=kind, status=desc.status, config=config)
    logger.warning("Unrecognized resource type '{}'.".format(kind))


@sentry
def create_index(
    name: str,
    wait: bool = True,
    engine_type: str = "approximated",
    metric: str = "cosine",
    shards: int = 1,
    replicas: int = 1,
    gateway_replicas: int = 1,
    engine_args: dict = None,
) -> Optional[Tuple[dict, ProgressBar]]:
    """Creates a Pinecone index.

    :param name: the name of the index.
    :type name: str
    :param wait: wait for the index to deploy. Defaults to ``True``
    :type wait: bool
    :param engine_type: type of engine, one of {"approximated", "exact"}, defaults to "approximated".
        The "approximated" engine uses fast approximate search algorithms developed by Pinecone.
        The "exact" engine uses accurate exact search algorithms.
        It performs exhaustive searches and thus it is usually slower than the "approximated" engine.
    :type engine_type: str, optional
    :param metric: type of metric used in the vector index, one of {"cosine", "dotproduct", "euclidean"}, defaults to "cosine".
        Use "cosine" for cosine similarity,
        "dotproduct" for dot-product,
        and "euclidean" for euclidean distance.
    :type metric: str, optional
    :param shards: the number of shards for the engine, defaults to 1.
        As a general guideline, use 1 shard per 1 GB of data.
    :type shards: int, optional
    :param replicas: the number of replicas, defaults to 1.
        Use at least 2 replicas if you need high availability (99.99% uptime) for querying.
        For additional throughput (QPS) your service needs to support, provision additional replicas.
    :type replicas: int, optional
    :param gateway_replicas: number of replicas of both the gateway and the aggregator.
    :type gateway_replicas: int
    :param engine_args: advanced arguments for the engine instance in the graph.
    :type engine_args: dict
    """
    return create(
        name=name,
        kind=ResourceType.INDEX.value,
        wait=wait,
        engine_type=engine_type,
        metric=metric,
        shards=shards,
        replicas=replicas,
        gateway_replicas=gateway_replicas,
        engine_args=engine_args,
    )


@sentry
def delete_index(name: str, wait: bool = True) -> Optional[Tuple[dict, ProgressBar]]:
    """Deletes a Pinecone index.

    :param name: the name of the index.
    :type name: str
    :param wait: wait for the index to deploy. Defaults to ``True``
    :type wait: bool
    """
    return delete(name=name, kind=ResourceType.INDEX.value, wait=wait)


@sentry
def list_indexes() -> Optional[List[str]]:
    """Lists all indexes."""
    return ls(kind=ResourceType.INDEX.value)


@sentry
def describe_index(name: str) -> Optional[ResourceDescription]:
    """Describes an index.

    :param name: the name of the index.
    :type name: str
    """
    return describe(name=name, kind=ResourceType.INDEX.value)


#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.network.zmq import SocketType
from pinecone.protos import core_pb2
from pinecone.utils import module_name, get_container_memory_usage
from pinecone.utils.constants import PERSISTENT_VOLUME_MOUNT, MEMORY_LIMIT_BYTES, NodeType
from pinecone.utils import logger

import os.path, os
import json
import argparse
import asyncio
from typing import Dict, Optional


class Function:
    """
    Base class for zmq-networked microservices
    """

    @classmethod
    def from_args(cls, args):
        """
        Create instance from cli args
        :param name:
        :param args:
        :return:
        """
        config = json.loads(args.config)
        node_type = NodeType(args.node_type) if args.node_type else NodeType.STANDARD
        if 'image' in config:
            image = config.pop('image')
            return cls(image, replicas=args.replicas, shards=args.shards, name=args.name, node_type=node_type, **config)
        return cls(replicas=args.replicas, shards=args.shards, name=args.name, node_type=node_type, **config)

    def to_args(self):
        """
        Return the command line args to run this instance
        :return:
        """
        return ['--config', json.dumps(self.config, separators=(',', ':')),
                '--replicas', str(self.replicas),
                '--shards', str(self.shards),
                '--name', str(self.name),
                '--node_type', self._node_type.value]

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser):
        parser.add_argument('--config', type=str, help='json config for functions')
        parser.add_argument('--replicas', type=int, help='total number of replicas', default='1')
        parser.add_argument('--shards', type=int, help='total number of shards', default='1')
        parser.add_argument('--name', type=str, help='functions name')
        parser.add_argument('--node_type', type=str, help='type of node to schedule function on', default=NodeType.STANDARD.value)

    def __init__(self,
                 in_sock_type: SocketType = SocketType.PULL,
                 out_sock_type: SocketType = SocketType.PUSH,
                 replicas: int = 1,
                 shards: int = 1,
                 name: str = None,
                 node_type: NodeType = NodeType.STANDARD,
                 **config):
        """
        :param config: kwargs must all be json-serializable settings for this functions
        """
        self.in_sock_type = in_sock_type
        self.out_sock_type = out_sock_type
        self.config = config
        self._replicas = replicas
        self._id = 0
        self._shards = shards
        self._name = name
        self._node_type = node_type

    @property
    def image(self):
        return 'pinecone/base'

    def get_config(self) -> dict:
        """
        Get the serializable runtime config required for this functions
        :return:
        """
        return self.config

    def setup(self):
        """
        Function that gets called prior to functions start, e.g. for downloading models
        :return:
        """

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        """Handles requests."""
        raise NotImplementedError

    @property
    def name(self):
        # TODO: deployment name should be normalized (regex used for validation is '[a-z0-9]([-a-z0-9]*[a-z0-9])?')
        if self._name is None:
            self._name = '-'.join([self.__class__.__name__.lower(), str(self.__hash__()).lower()])

        return self._name

    @property
    def replicas(self):
        return self._replicas

    @property
    def shards(self):
        return self._shards

    @property 
    def id(self):
        return self._id
    
    @id.setter
    def id(self, value: int):
        self._id = value

    @property
    def stats_frequency_seconds(self):
        return 0

    def get_stats(self) -> Dict:
        return {}

    async def export_stats(self):
        if self.stats_frequency_seconds == 0:
            return
        while True:
            await asyncio.sleep(self.stats_frequency_seconds)
            stats = self.get_stats()
            logger.opt(raw=True).info(json.dumps({"recordType": "functionStats", "name": self.name, **stats}) + "\n")

    def to_dict(self):
        return {
            'name': self.name,
            'class': module_name(self),
            'args': self.to_args()
        }

    @property
    def memory_request(self) -> int:
        """
        :return: Requested memory resources in mb
        """
        return 300

    @property
    def cpu_request(self) -> int:
        """
        :return: Request CPU in 1/1000th of a core
        """
        return 300

    @property
    def volume_request(self) -> Optional[int]:
        """
        Volume request in GB or None for ephemeral (stateless) functions
        :return:
        """
        return None

    @property
    def default_persistent_dir(self):
        persistent_mount = os.path.exists(PERSISTENT_VOLUME_MOUNT) and os.access(PERSISTENT_VOLUME_MOUNT, os.W_OK)
        data_dir = PERSISTENT_VOLUME_MOUNT if persistent_mount else '/tmp/data'
        persistent_dir = os.path.join(data_dir, os.environ.get('HOSTNAME'))
        os.makedirs(persistent_dir, exist_ok=True)
        return os.path.join(persistent_dir, self.name)

    @property
    def full(self) -> bool:
        if MEMORY_LIMIT_BYTES == 0:
            return False
        return get_container_memory_usage() > MEMORY_LIMIT_BYTES

    def cleanup(self):
        pass



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Iterable

from pinecone.functions.model.scorer import Scorer
from pinecone import utils

import numpy as np


@utils.hubify
class DotScorer(Scorer):

    def score(self,
              data_queries: Iterable[np.ndarray],
              data_items: Iterable[Iterable[np.ndarray]]
              ) -> Iterable[Iterable[float]]:
        return [list(np.dot(queries, np.transpose(items)).tolist()) for queries, items in zip(data_queries, data_items)]
